"""Fsspec filesystem classes with UPath integration."""

from __future__ import annotations

from fsspec.implementations.cached import SimpleCacheFileSystem as _SimpleCacheFileSystem
from fsspec.implementations.data import DataFileSystem as _DataFileSystem
from fsspec.implementations.github import GithubFileSystem as _GithubFileSystem

# from fsspec.implementations.http import HTTPFileSystem as _HTTPFileSystem
# from fsspec.implementations.local import LocalFileSystem as _LocalFileSystem
# from fsspec.implementations.memory import MemoryFileSystem as _MemoryFileSystem
from fsspec.implementations.tar import TarFileSystem as _TarFileSystem
from fsspec.implementations.zip import ZipFileSystem as _ZipFileSystem
from upath.implementations.cached import SimpleCachePath
from upath.implementations.data import DataPath
from upath.implementations.github import GitHubPath

# from upath.implementations.http import HTTPPath
# from upath.implementations.local import FilePath
# from upath.implementations.memory import MemoryPath
from upath.implementations.tar import TarPath
from upath.implementations.zip import ZipPath

from upathtools.filesystems.base import BaseFileSystem


# class LocalFileSystem(BaseFileSystem[FilePath], _LocalFileSystem):
#     """Local filesystem with UPath integration."""

#     upath_cls = FilePath


# class MemoryFileSystem(BaseFileSystem[MemoryPath], _MemoryFileSystem):
#     """In-memory filesystem with UPath integration."""

#     upath_cls = MemoryPath


# class HTTPFileSystem(BaseFileSystem[HTTPPath], _HTTPFileSystem):
#     """HTTP/HTTPS filesystem with UPath integration."""

#     upath_cls = HTTPPath


class GithubFileSystem(BaseFileSystem[GitHubPath], _GithubFileSystem):
    """GitHub filesystem with UPath integration."""

    upath_cls = GitHubPath


class TarFileSystem(BaseFileSystem[TarPath], _TarFileSystem):
    """Tar archive filesystem with UPath integration."""

    upath_cls = TarPath


class ZipFileSystem(BaseFileSystem[ZipPath], _ZipFileSystem):
    """Zip archive filesystem with UPath integration."""

    upath_cls = ZipPath


class SimpleCacheFileSystem(BaseFileSystem[SimpleCachePath], _SimpleCacheFileSystem):
    """Simple cache filesystem with UPath integration."""

    upath_cls = SimpleCachePath


class DataFileSystem(BaseFileSystem[DataPath], _DataFileSystem):
    """Data URI filesystem with UPath integration."""

    upath_cls = DataPath



"""Microsandbox async filesystem implementation for upathtools."""

from __future__ import annotations

import logging
from typing import TYPE_CHECKING, Any, Literal, Required, overload

from upathtools.async_helpers import sync_wrapper
from upathtools.filesystems.base import BaseAsyncFileSystem, BaseUPath, FileInfo


if TYPE_CHECKING:
    from microsandbox import BaseSandbox


class MicrosandboxInfo(FileInfo, total=False):
    """Info dict for Microsandbox filesystem paths."""

    size: Required[int]


logger = logging.getLogger(__name__)


class MicrosandboxPath(BaseUPath[MicrosandboxInfo]):
    """Microsandbox-specific UPath implementation."""

    __slots__ = ()


class MicrosandboxFS(BaseAsyncFileSystem[MicrosandboxPath, MicrosandboxInfo]):
    """Async filesystem for Microsandbox environments.

    This filesystem provides access to files within a Microsandbox environment,
    allowing you to read, write, and manipulate files remotely through the
    Microsandbox CLI interface.
    """

    protocol = "microsandbox"
    upath_cls = MicrosandboxPath
    root_marker = "/"
    cachable = False  # Disable fsspec caching to prevent instance sharing

    def __init__(
        self,
        sandbox: BaseSandbox | None = None,
        server_url: str | None = None,
        namespace: str = "default",
        name: str | None = None,
        api_key: str | None = None,
        image: str | None = None,
        memory: int = 512,
        cpus: float = 1.0,
        **kwargs: Any,
    ) -> None:
        """Initialize Microsandbox filesystem.

        Args:
            sandbox: Existing sandbox instance to use
            server_url: Microsandbox server URL
            namespace: Sandbox namespace
            name: Sandbox name
            api_key: API key for authentication
            image: Docker image to use
            memory: Memory limit in MB
            cpus: CPU limit
            **kwargs: Additional filesystem arguments
        """
        super().__init__(**kwargs)
        self._sandbox = sandbox
        self.server_url = server_url
        self.namespace = namespace
        self.name = name
        self.api_key = api_key
        self.image = image
        self.memory = memory
        self.cpus = cpus

    async def _get_sandbox(self) -> BaseSandbox:
        """Get sandbox instance."""
        if self._sandbox is None:
            msg = "No sandbox provided and cannot create one without sandbox class"
            raise RuntimeError(msg)

        if not self._sandbox._is_started:
            await self._sandbox.start(
                image=self.image,
                memory=self.memory,
                cpus=self.cpus,
            )
            logger.info("Started Microsandbox: %s", self._sandbox._name)

        return self._sandbox

    async def close_session(self) -> None:
        """Close sandbox session."""
        if self._sandbox is not None and self._sandbox._is_started:
            await self._sandbox.stop()
            logger.info("Stopped Microsandbox: %s", self._sandbox._name)

    @overload
    async def _ls(
        self,
        path: str,
        detail: Literal[True] = ...,
        **kwargs: Any,
    ) -> list[MicrosandboxInfo]: ...

    @overload
    async def _ls(self, path: str, detail: bool, **kwargs: Any) -> list[str]: ...

    async def _ls(
        self,
        path: str,
        detail: bool = True,
        **kwargs: Any,
    ) -> list[str] | list[MicrosandboxInfo]:
        """List directory contents."""
        sandbox = await self._get_sandbox()
        # Use ls -la to get detailed directory listing
        result = await sandbox.command.run("ls", ["-la", path])
        output = await result.output()
        stderr = await result.error()
        if result.exit_code != 0:
            if "No such file or directory" in stderr:
                msg = f"Path not found: {path}"
                raise FileNotFoundError(msg)
            msg = f"Failed to list directory {path}: {stderr}"
            raise OSError(msg)

        files = []
        for line in output.strip().split("\n"):
            if not line or line.startswith("total"):
                continue

            parts = line.split()
            min_parts = 9
            if len(parts) < min_parts:
                continue
            permissions = parts[0]
            name = parts[-1]
            if name in (".", ".."):
                continue
            is_dir = permissions.startswith("d")
            full_path = f"{path.rstrip('/')}/{name}" if path != "/" else f"/{name}"

            files.append(
                MicrosandboxInfo(
                    name=full_path,
                    size=0 if is_dir else int(parts[4]) if parts[4].isdigit() else 0,
                    type="directory" if is_dir else "file",
                )
            )
        return files if detail else [f["name"] for f in files]  # type: ignore[misc]

    async def _cat_file(
        self,
        path: str,
        start: int | None = None,
        end: int | None = None,
        **kwargs: Any,
    ) -> bytes:
        """Read file contents using cat command."""
        sandbox = await self._get_sandbox()
        result = await sandbox.command.run("cat", [path])
        stdout = await result.output()
        stderr = await result.error()
        if result.exit_code != 0:
            if "No such file or directory" in stderr:
                msg = f"File not found: {path}"
                raise FileNotFoundError(msg)
            msg = f"Failed to read file {path}: {stderr}"
            raise OSError(msg)

        content = stdout.encode("utf-8")
        if start is not None or end is not None:
            return content[start:end]
        return content

    async def _pipe_file(self, path: str, value: bytes, **kwargs: Any) -> None:
        """Write file contents using shell redirection."""
        sandbox = await self._get_sandbox()
        parent = path.rsplit("/", 1)[0]
        if parent and parent != path:
            await self._mkdir(parent, create_parents=True)
        # Use echo with redirection to write content
        # Use printf to handle special characters and newlines properly
        result = await sandbox.command.run(
            "sh", ["-c", f"printf '%s' '{value.decode('utf-8')}' > '{path}'"]
        )
        stderr = await result.error()
        if result.exit_code != 0:
            msg = f"Failed to write file {path}: {stderr}"
            raise OSError(msg)

    async def _mkdir(self, path: str, create_parents: bool = True, **kwargs: Any) -> None:
        """Create directory using mkdir command."""
        sandbox = await self._get_sandbox()

        args = [path]
        if create_parents:
            args = ["-p", *args]

        result = await sandbox.command.run("mkdir", args)
        stderr = await result.error()
        if result.exit_code != 0:
            msg = f"Failed to create directory {path}: {stderr}"
            raise OSError(msg)

    async def _rm_file(self, path: str, **kwargs: Any) -> None:
        """Remove file using rm command."""
        sandbox = await self._get_sandbox()
        result = await sandbox.command.run("rm", ["-f", path])
        stderr = await result.error()
        if result.exit_code != 0:
            msg = f"Failed to remove file {path}: {stderr}"
            raise OSError(msg)

    async def _rmdir(self, path: str, **kwargs: Any) -> None:
        """Remove directory using rmdir command."""
        sandbox = await self._get_sandbox()
        result = await sandbox.command.run("rmdir", [path])
        stderr = await result.error()
        if result.exit_code != 0:
            msg = f"Failed to remove directory {path}: {stderr}"
            raise OSError(msg)

    async def _exists(self, path: str, **kwargs: Any) -> bool:
        """Check if path exists using test command."""
        sandbox = await self._get_sandbox()
        result = await sandbox.command.run("test", ["-e", path])
        return result.exit_code == 0

    async def _isfile(self, path: str, **kwargs: Any) -> bool:
        """Check if path is a file using test command."""
        sandbox = await self._get_sandbox()
        result = await sandbox.command.run("test", ["-f", path])
        return result.exit_code == 0

    async def _isdir(self, path: str, **kwargs: Any) -> bool:
        """Check if path is a directory using test command."""
        sandbox = await self._get_sandbox()
        result = await sandbox.command.run("test", ["-d", path])
        return result.exit_code == 0

    async def _size(self, path: str, **kwargs: Any) -> int:
        """Get file size using stat command."""
        sandbox = await self._get_sandbox()
        result = await sandbox.command.run("stat", ["-c", "%s", path])
        stdout = await result.output()
        stderr = await result.error()
        if result.exit_code != 0:
            if "No such file or directory" in stderr:
                msg = f"File not found: {path}"
                raise FileNotFoundError(msg)
            msg = f"Failed to get size for {path}: {stderr}"
            raise OSError(msg)

        return int(stdout.strip())

    async def _modified(self, path: str, **kwargs: Any) -> float:
        """Get file modification time using stat command."""
        sandbox = await self._get_sandbox()
        result = await sandbox.command.run("stat", ["-c", "%Y", path])
        stdout = await result.output()
        stderr = await result.error()
        if result.exit_code != 0:
            if "No such file or directory" in stderr:
                msg = f"File not found: {path}"
                raise FileNotFoundError(msg)
            msg = f"Failed to get modification time for {path}: {stderr}"
            raise OSError(msg)

        return float(stdout.strip())

    # Sync wrapper methods
    ls = sync_wrapper(_ls)  # pyright: ignore[reportAssignmentType]
    cat_file = sync_wrapper(_cat_file)  # pyright: ignore[reportAssignmentType]
    pipe_file = sync_wrapper(_pipe_file)  # pyright: ignore[reportAssignmentType]
    mkdir = sync_wrapper(_mkdir)
    rm_file = sync_wrapper(_rm_file)
    rmdir = sync_wrapper(_rmdir)
    exists = sync_wrapper(_exists)  # pyright: ignore[reportAssignmentType]
    isfile = sync_wrapper(_isfile)
    isdir = sync_wrapper(_isdir)
    size = sync_wrapper(_size)
    modified = sync_wrapper(_modified)

    def cat(self, path: str, **kwargs: Any) -> bytes:
        """Read file contents (sync wrapper)."""
        result = self.cat_file(path, **kwargs)
        if isinstance(result, str):
            return result.encode("utf-8")
        return result

    def info(self, path: str, **kwargs: Any) -> MicrosandboxInfo:
        """Get file info (sync wrapper)."""
        return MicrosandboxInfo(
            name=path,
            size=self.size(path) if self.isfile(path) else 0,  # pyright: ignore[reportArgumentType]
            type="directory" if self.isdir(path) else "file",
        )



"""Async local filesystem with ripgrep-rs optimizations."""

from __future__ import annotations

from asyncio import get_running_loop
from functools import partial
from inspect import iscoroutinefunction
import json
import os
from pathlib import Path
import shutil
from typing import Any, Literal, Required, overload

from fsspec.implementations.local import LocalFileSystem

from upathtools.async_helpers import wrap
from upathtools.filesystems.base import BaseAsyncFileSystem, BaseUPath, GrepMatch
from upathtools.filesystems.base.file_objects import FileInfo


class LocalFileInfo(FileInfo, total=False):
    """Info dict for local filesystem paths."""

    size: int
    created: float
    islink: bool
    mode: int
    uid: int
    gid: int
    mtime: float
    ino: int
    nlink: int
    destination: Required[bool]


# Windows path normalization for fsspec compatibility
_IS_WINDOWS = os.name == "nt"


def _normalize_path(path: str) -> str:
    """Normalize path separators to forward slashes (fsspec convention)."""
    return path.replace("\\", "/") if _IS_WINDOWS else path


class LocalPath(BaseUPath[LocalFileInfo]):
    """UPath implementation for local filesystem."""

    __slots__ = ()


class AsyncLocalFileSystem(BaseAsyncFileSystem[LocalPath, LocalFileInfo], LocalFileSystem):
    """Async implementation of LocalFileSystem.

    This filesystem provides both async and sync methods. The sync methods are not
    overridden and use LocalFileSystem's implementation.

    The async methods run the respective sync methods in a threadpool executor.
    It also provides open_async() method that supports asynchronous file operations,
    using `aiofile`_.

    Note that some async methods like _find may call these wrapped async methods
    many times, and might have high overhead.
    In that case, it might be faster to run the whole operation in a threadpool,
    which is available as `_*_async()` versions of the API.
    eg: _find_async()/_get_file_async, etc.

    .. aiofile:
        https://github.com/mosquito/aiofile
    """

    mirror_sync_methods = False
    upath_cls = LocalPath

    _cat_file = wrap(LocalFileSystem.cat_file)  # type: ignore[assignment]
    _chmod = wrap(LocalFileSystem.chmod)
    _cp_file = wrap(LocalFileSystem.cp_file)  # type: ignore[assignment]
    _created = wrap(LocalFileSystem.created)
    _find_async = wrap(LocalFileSystem.find)
    _get_file_async = wrap(LocalFileSystem.get_file)
    _islink = wrap(LocalFileSystem.islink)
    _lexists = wrap(LocalFileSystem.lexists)
    _link = wrap(LocalFileSystem.link)
    _makedirs = wrap(LocalFileSystem.makedirs)  # type: ignore[assignment]
    _mkdir = wrap(LocalFileSystem.mkdir)  # type: ignore[assignment]
    _modified = wrap(LocalFileSystem.modified)
    # `mv_file` was renamed to `mv` in fsspec==2024.5.0
    # https://github.com/fsspec/filesystem_spec/pull/1585
    _mv = wrap(getattr(LocalFileSystem, "mv", None) or LocalFileSystem.mv_file)  # type: ignore[arg-type,assignment]
    _mv_file = _mv  # type: ignore[assignment]
    _pipe_file = wrap(LocalFileSystem.pipe_file)  # type: ignore[assignment]
    _put_file = wrap(LocalFileSystem.put_file)  # type: ignore[assignment]
    _read_bytes = wrap(LocalFileSystem.read_bytes)
    _read_text = wrap(LocalFileSystem.read_text)
    _rm = wrap(LocalFileSystem.rm)  # type: ignore[assignment]
    _rm_file = wrap(LocalFileSystem.rm_file)  # type: ignore[assignment]
    _rmdir = wrap(LocalFileSystem.rmdir)
    _touch = wrap(LocalFileSystem.touch)
    _symlink = wrap(LocalFileSystem.symlink)
    _write_bytes = wrap(LocalFileSystem.write_bytes)
    _write_text = wrap(LocalFileSystem.write_text)

    async def _info(self, path: str, **kwargs: Any) -> LocalFileInfo:
        """Get info for a single path."""
        loop = get_running_loop()
        return await loop.run_in_executor(None, partial(LocalFileSystem.info, self, path, **kwargs))  # type: ignore[return-value]

    @overload
    async def _ls(
        self,
        path: str,
        detail: Literal[True] = ...,
        **kwargs: Any,
    ) -> list[LocalFileInfo]: ...

    @overload
    async def _ls(self, path: str, detail: bool, **kwargs: Any) -> list[str]: ...

    async def _ls(
        self,
        path: str,
        detail: bool = True,
        **kwargs: Any,
    ) -> list[LocalFileInfo] | list[str]:
        """List directory contents."""
        loop = get_running_loop()
        return await loop.run_in_executor(  # type: ignore[return-value]
            None, partial(LocalFileSystem.ls, self, path, detail=detail, **kwargs)
        )

    async def _get_file(
        self,
        src: str,
        dst: Any,
        **kwargs: Any,
    ) -> None:
        if not iscoroutinefunction(getattr(dst, "write", None)):
            src = self._strip_protocol(src)
            return await self._get_file_async(src, dst)

        async with await self.open_async(src, "rb") as fsrc:
            while True:
                buf = await fsrc.read(length=shutil.COPY_BUFSIZE)  # type: ignore[attr-defined]
                if not buf:
                    break
                await dst.write(buf)

    async def open_async(
        self,
        path: str,
        mode: str = "rb",
        **kwargs: Any,
    ) -> Any:
        import aiofile

        path = self._strip_protocol(path)
        if self.auto_mkdir and "w" in mode:
            await self._makedirs(self._parent(path), exist_ok=True)
        return await aiofile.async_open(path, mode, **kwargs)

    # -------------------------------------------------------------------------
    # ripgrep-rs optimized methods (optional fast path when available)
    # -------------------------------------------------------------------------

    @overload
    async def _find(
        self,
        path: str,
        maxdepth: int | None = None,
        withdirs: bool = False,
        *,
        detail: Literal[False] = False,
        **kwargs: Any,
    ) -> list[str]: ...

    @overload
    async def _find(
        self,
        path: str,
        maxdepth: int | None = None,
        withdirs: bool = False,
        *,
        detail: Literal[True],
        **kwargs: Any,
    ) -> dict[str, LocalFileInfo]: ...

    async def _find(
        self,
        path: str,
        maxdepth: int | None = None,
        withdirs: bool = False,
        *,
        detail: bool = False,
        **kwargs: Any,
    ) -> list[str] | dict[str, LocalFileInfo]:
        """Find files recursively, using ripgrep-rs for speed when available.

        Uses ripgrep-rs for both detail=False and detail=True modes.
        """
        from ripgrep_rs import files as rg_files, files_with_info

        loop = get_running_loop()
        abs_path = str(Path(self._strip_protocol(path)).resolve())

        if detail:
            # Use files_with_info for metadata - returns dict[str, FileInfo]
            rg_result = await loop.run_in_executor(
                None,
                partial(
                    files_with_info,
                    paths=[abs_path],
                    hidden=kwargs.get("hidden", False),
                    no_ignore=kwargs.get("no_ignore", False),
                    include_dirs=withdirs,
                    absolute=True,
                ),
            )
            # Convert FileInfo objects to fsspec-compatible dicts
            # Normalize paths to forward slashes (fsspec convention)
            return {
                _normalize_path(p): LocalFileInfo(
                    name=info.name,
                    size=info.size,
                    type="directory" if info.type == "directory" else "file",
                    created=info.created,
                    islink=info.islink,
                    mode=info.mode,
                    uid=info.uid,
                    gid=info.gid,
                    mtime=info.mtime,
                    ino=info.ino,
                    nlink=info.nlink,
                    destination=False,  # Not a symlink destination
                )
                for p, info in rg_result.items()
            }

        # ripgrep-rs runs in a thread pool since it releases the GIL
        results = await loop.run_in_executor(
            None,
            partial(
                rg_files,
                paths=[abs_path],
                hidden=kwargs.get("hidden", False),
                no_ignore=kwargs.get("no_ignore", False),
                include_dirs=withdirs,
                absolute=True,  # Ensure absolute paths (for Windows)
            ),
        )
        # Normalize paths to forward slashes (fsspec convention)
        return [_normalize_path(p) for p in results]

    @overload
    async def _glob(
        self,
        path: str,
        maxdepth: int | None = None,
        *,
        detail: Literal[False] = False,
        **kwargs: Any,
    ) -> list[str]: ...

    @overload
    async def _glob(
        self,
        path: str,
        maxdepth: int | None = None,
        *,
        detail: Literal[True],
        **kwargs: Any,
    ) -> dict[str, LocalFileInfo]: ...

    async def _glob(
        self,
        path: str,
        maxdepth: int | None = None,
        *,
        detail: bool = False,
        **kwargs: Any,
    ) -> list[str] | dict[str, LocalFileInfo]:
        """Glob for files, using ripgrep-rs for speed when available.

        Uses ripgrep-rs for both detail=False and detail=True modes.
        Falls back to base implementation for complex glob patterns that ripgrep can't handle.
        """
        from glob import has_magic

        from ripgrep_rs import files as rg_files, files_with_info

        stripped = self._strip_protocol(path)
        if not has_magic(path):
            # No glob pattern - fall back to base implementation
            if detail:
                return await super()._glob(path, maxdepth=maxdepth, detail=True, **kwargs)
            return await super()._glob(path, maxdepth=maxdepth, detail=False, **kwargs)

        # Fast path: use ripgrep-rs for globs
        loop = get_running_loop()

        if "**" in stripped:
            # Recursive glob - split at first **
            # e.g., "/home/user/**/*.py" -> base="/home/user", pattern="**/*.py"
            idx = stripped.find("**")
            base = stripped[:idx].rstrip("/") or "."
            glob_pattern = stripped[idx:]
            depth = None  # No depth limit for ** patterns
        else:
            # Non-recursive glob - use max_depth=1
            # e.g., "/home/user/*.py" -> base="/home/user", pattern="*.py"
            p = Path(stripped)
            base = str(p.parent) if str(p.parent) != "." else "."
            glob_pattern = p.name
            depth = 1  # Only match in the specified directory

        abs_base = str(Path(base).resolve())
        if detail:
            # Use files_with_info for metadata - returns dict[str, FileInfo]
            rg_result = await loop.run_in_executor(
                None,
                partial(
                    files_with_info,
                    paths=[abs_base],
                    globs=[glob_pattern],
                    hidden=kwargs.get("hidden", False),
                    no_ignore=kwargs.get("no_ignore", False),
                    max_depth=depth,
                    absolute=True,
                ),
            )
            # Convert FileInfo objects to fsspec-compatible dicts
            # Normalize paths to forward slashes (fsspec convention)
            return {
                _normalize_path(p): LocalFileInfo(
                    name=info.name,
                    size=info.size,
                    type="directory" if info.type == "directory" else "file",
                    created=info.created,
                    islink=info.islink,
                    mode=info.mode,
                    uid=info.uid,
                    gid=info.gid,
                    mtime=info.mtime,
                    ino=info.ino,
                    nlink=info.nlink,
                    destination=False,  # Not a symlink destination
                )
                for p, info in rg_result.items()
            }
        results = await loop.run_in_executor(
            None,
            partial(
                rg_files,
                paths=[abs_base],
                globs=[glob_pattern],
                hidden=kwargs.get("hidden", False),
                no_ignore=kwargs.get("no_ignore", False),
                max_depth=depth,
                absolute=True,  # Ensure absolute paths (for Windows)
            ),
        )
        # Normalize paths to forward slashes (fsspec convention)
        return [_normalize_path(p) for p in results]

    async def _grep(
        self,
        path: str,
        pattern: str,
        *,
        max_count: int | None = None,
        case_sensitive: bool | None = None,
        hidden: bool = False,
        no_ignore: bool = False,
        globs: list[str] | None = None,
        context_before: int | None = None,
        context_after: int | None = None,
        multiline: bool = False,
    ) -> list[GrepMatch]:
        """Search for pattern in files using ripgrep-rs.

        This provides a fast grep implementation using ripgrep's algorithms
        via the ripgrep-rs Python bindings.

        Args:
            path: Directory to search in.
            pattern: Regex pattern to search for.
            max_count: Maximum total number of matches to return.
            case_sensitive: Force case sensitivity (None = smart case).
            hidden: Search hidden files/directories.
            no_ignore: Don't respect .gitignore rules.
            globs: File patterns to include/exclude (e.g., ``['*.py', '!*_test.py']``).
            context_before: Lines of context before match.
            context_after: Lines of context after match.
            multiline: Enable multiline matching.

        Returns:
            List of GrepMatch objects with file, line number, and matched text.
        """
        from ripgrep_rs import search as rg_search

        abs_path = str(Path(self._strip_protocol(path)).resolve())

        # Build kwargs for ripgrep
        rg_kwargs: dict[str, Any] = {
            "patterns": [pattern],
            "paths": [abs_path],
            "hidden": hidden,
            "no_ignore": no_ignore,
            "json": True,  # Get structured output
        }
        if max_count is not None:
            rg_kwargs["max_count"] = max_count
        if case_sensitive is not None:
            rg_kwargs["case_sensitive"] = case_sensitive
        if globs:
            rg_kwargs["globs"] = globs
        if context_before is not None:
            rg_kwargs["before_context"] = context_before
        if context_after is not None:
            rg_kwargs["after_context"] = context_after
        if multiline:
            rg_kwargs["multiline"] = True

        # Run ripgrep in thread pool (ripgrep-rs releases the GIL)
        raw_results = await wrap(rg_search)(**rg_kwargs)
        # Parse JSON Lines output into GrepMatch objects
        matches: list[GrepMatch] = []
        for file_result in raw_results:
            for line in file_result.strip().split("\n"):
                if not line:
                    continue
                try:
                    obj = json.loads(line)
                    if obj.get("type") == "match":
                        data = obj["data"]
                        # Normalize path to forward slashes (fsspec convention)
                        match_path = _normalize_path(data["path"]["text"])
                        matches.append(
                            GrepMatch(
                                path=match_path,
                                line_number=data["line_number"],
                                text=data["lines"]["text"].rstrip("\n"),
                                submatches=[
                                    (sm["start"], sm["end"]) for sm in data.get("submatches", [])
                                ],
                                absolute_offset=data.get("absolute_offset", 0),
                            )
                        )
                except (json.JSONDecodeError, KeyError):
                    continue

        return matches


def register_async_local_fs() -> bool:
    """Register AsyncLocalFileSystem in fsspec registry, replacing morefs.

    This registers our optimized AsyncLocalFileSystem (with ripgrep-rs support)
    as the handler for the 'asynclocal' protocol, overriding the default morefs
    implementation.

    Returns:
        True if registration succeeded, False if upath is not available.
    """
    import fsspec
    from fsspec.implementations.local import make_path_posix
    from fsspec.utils import stringify_path
    from upath._flavour_sources import AbstractFileSystemFlavour
    from upath.registry import register_implementation

    # Register in fsspec registry (clobber=True to override morefs)
    fsspec.register_implementation(
        "asynclocal",
        "upathtools.filesystems.async_local_fs.AsyncLocalFileSystem",
        clobber=True,
    )

    class AsyncLocalFileSystemFlavour(AbstractFileSystemFlavour):
        __orig_class__ = "upathtools.filesystems.async_local.AsyncLocalFileSystem"
        protocol = ()
        root_marker = "/"
        sep = "/"
        local_file = True

        @classmethod
        def _strip_protocol(cls, path):  # type: ignore[override]
            path = stringify_path(path)
            if path.startswith("file://"):
                path = path[7:]
            elif path.startswith("file:"):
                path = path[5:]
            elif path.startswith("local://"):
                path = path[8:]
            elif path.startswith("local:"):
                path = path[6:]

            path = str(make_path_posix(path))
            if os.sep != "/":
                if path[1:2] == ":":
                    drive, path = path[:2], path[2:]
                elif path[:2] == "//":
                    if (index1 := path.find("/", 2)) == -1 or (
                        index2 := path.find("/", index1 + 1)
                    ) == -1:
                        drive, path = path, ""
                    else:
                        drive, path = path[:index2], path[index2:]
                else:
                    drive = ""

                return drive + (path.rstrip("/") or cls.root_marker)

            return path.rstrip("/") or cls.root_marker

        @classmethod
        def _parent(cls, path):  # type: ignore[override]
            path = cls._strip_protocol(path)
            if os.sep == "/":
                return path.rsplit("/", 1)[0] or "/"
            path_ = path.rsplit("/", 1)[0]
            if len(path_) <= 3 and path_[1:2] == ":":  # noqa: PLR2004
                return path_[0] + ":/"
            return path_

    register_implementation("asynclocal", LocalPath, clobber=True)
    return True


# Keep old name for backwards compatibility
register_flavour = register_async_local_fs


if __name__ == "__main__":
    import asyncio

    async def main():
        fs = AsyncLocalFileSystem()
        await fs._mkdir("test")
        ls = await fs._ls("")
        print(ls)

    asyncio.run(main())



"""CLI command parser and executor for filesystem operations.

This module provides a shell-like interface for executing commands on filesystems and paths.
"""

from __future__ import annotations

import asyncio
from collections.abc import Sequence
from dataclasses import dataclass
import shlex
from typing import TYPE_CHECKING, Any


if TYPE_CHECKING:
    from upath import UPath

SHORT_TO_LONG = {
    "r": "recursive",
    "i": "ignore_case",
    "v": "invert_match",
    "w": "whole_word",
    "F": "fixed_string",
    "m": "max_count",
    "B": "context_before",
    "A": "context_after",
    "n": "n",
    "a": "all_",
    "l": "long",
    "h": "human_readable",
    "s": "summarize",
    "d": "max_depth",
    "f": "force",
    "p": "parents",
}


@dataclass
class CLIResult[TData]:
    """Result from CLI command execution."""

    data: TData
    command: str

    def __str__(self) -> str:
        """Format result as string."""
        if isinstance(self.data, list):
            return "\n".join(str(item) for item in self.data)
        return str(self.data)

    def __iter__(self):
        """Allow iteration over results."""
        if isinstance(self.data, Sequence):
            yield from self.data
        else:
            yield self.data


async def execute_cli_async(command: str, base: UPath) -> CLIResult:  # noqa: PLR0911
    """Execute a CLI-style command on a filesystem/path asynchronously.

    Args:
        command: Shell-like command string (e.g., "grep pattern file.txt -i")
        base: Base UPath to execute command relative to

    Returns:
        CLIResult with command output

    Examples:
        >>> path = UPath(".")
        >>> result = await execute_cli_async("grep TODO *.py -r", path)
        >>> for match in result:
        ...     print(match)
    """
    from upathtools.cli_ops import acat, adiff, adu, afind, agrep, ahead, als, atail, awc

    parts = shlex.split(command)
    if not parts:
        msg = "Empty command"
        raise ValueError(msg)

    cmd = parts[0]
    args, kwargs = _parse_args(parts[1:])

    match cmd:
        case "grep":
            if len(args) < 1:
                msg = "grep requires pattern argument"
                raise ValueError(msg)

            path = args[1] if len(args) > 1 else "."
            grep_results = [result async for result in agrep(cmd, path, base, **kwargs)]
            return CLIResult(grep_results, f"grep {cmd} {path}")
        case "find":
            path = args[0] if args else "."

            # Map common find args
            if "name" in args:
                idx = args.index("name")
                if idx + 1 < len(args):
                    kwargs["name"] = args[idx + 1]

            find_results = [i async for i in afind(path, base, **kwargs)]
            return CLIResult(find_results, f"find {path}")
        case "head":
            if not args:
                msg = "head requires file argument"
                raise ValueError(msg)

            result = await ahead(cmd, base, **kwargs)
            return CLIResult(result, f"head {cmd}")
        case "tail":
            if not args:
                msg = "tail requires file argument"
                raise ValueError(msg)

            result = await atail(cmd, base, **kwargs)
            return CLIResult(result, f"tail {cmd}")

        case "cat":
            if not args:
                msg = "cat requires file argument(s)"
                raise ValueError(msg)

            result = await acat(*args, base=base, **kwargs)
            return CLIResult(result, f"cat {' '.join(args)}")

        case "wc":
            if not args:
                msg = "wc requires file argument"
                raise ValueError(msg)

            result = await awc(cmd, base, **kwargs)
            return CLIResult(result, f"wc {cmd}")

        case "ls":
            path = cmd if args else "."
            ls_results = await als(path, base, **kwargs)
            return CLIResult(ls_results, f"ls {path}")
        case "du":
            path = cmd if args else "."
            du_results = await adu(path, base, **kwargs)
            return CLIResult(du_results, f"du {path}")
        case "diff":
            if len(args) < 2:  # noqa: PLR2004
                msg = "diff requires two file arguments"
                raise ValueError(msg)

            result = await adiff(cmd, args[1], base, **kwargs)
            return CLIResult(result, f"diff {cmd} {args[1]}")
        case _:
            msg = f"Unknown command: {cmd}"
            raise ValueError(msg)


def execute_cli(command: str, base: UPath) -> CLIResult:
    """Execute a CLI-style command on a filesystem/path (sync wrapper)."""
    return asyncio.run(execute_cli_async(command, base))


def _parse_args(args: list[str]) -> tuple[list[str], dict[str, Any]]:
    """Parse positional args and flags from command arguments.

    Args:
        args: List of argument strings

    Returns:
        Tuple of (positional_args, kwargs_dict)
    """
    positional: list[str] = []
    kwargs: dict[str, Any] = {}
    i = 0

    while i < len(args):
        arg = args[i]

        if arg.startswith("--"):
            # Long option
            key = arg[2:].replace("-", "_")
            if "=" in key:
                k, v = key.split("=", 1)
                kwargs[k] = _parse_value(v)
            elif i + 1 < len(args) and not args[i + 1].startswith("-"):
                kwargs[key] = _parse_value(args[i + 1])
                i += 1
            else:
                kwargs[key] = True
        elif arg.startswith("-") and len(arg) > 1:
            # Short options
            for char in arg[1:]:
                flag_name = SHORT_TO_LONG.get(char, char)
                # Check if next arg is a value for this flag
                if char in "nBAmdc" and i + 1 < len(args) and not args[i + 1].startswith("-"):
                    kwargs[flag_name] = _parse_value(args[i + 1])
                    i += 1
                else:
                    kwargs[flag_name] = True
        else:
            positional.append(arg)

        i += 1

    return positional, kwargs


def _parse_value(value: str) -> Any:
    """Parse a string value to appropriate type."""
    if value.lower() == "true":
        return True
    if value.lower() == "false":
        return False
    try:
        return int(value)
    except ValueError:
        pass
    try:
        return float(value)
    except ValueError:
        pass
    return value



from __future__ import annotations

import asyncio
from asyncio import get_running_loop
import asyncio.events
import functools
from functools import partial, wraps
import threading
from typing import TYPE_CHECKING, Any

from fsspec.exceptions import FSTimeoutError


if TYPE_CHECKING:
    from collections.abc import Awaitable, Callable, Coroutine


async def _runner(
    event: threading.Event,
    coro: Coroutine[Any, Any, Any],
    result: list[Any],
    timeout: float | None = None,
) -> None:
    timeout = timeout if timeout else None  # convert 0 or 0.0 to None
    awaitable: Awaitable[Any] = coro
    if timeout is not None:
        awaitable = asyncio.wait_for(coro, timeout=timeout)
    try:
        result[0] = await awaitable
    except Exception as ex:  # noqa: BLE001
        result[0] = ex
    finally:
        event.set()


def sync[T](
    loop: asyncio.AbstractEventLoop,
    func: Callable[..., Coroutine[Any, Any, T]],
    *args: Any,
    timeout: float | None = None,
    **kwargs: Any,
) -> T:
    """Make loop run coroutine until it returns. Runs in other thread.

    Examples
    --------
    >>> fsspec.asyn.sync(fsspec.asyn.get_loop(), func, *args,
                         timeout=timeout, **kwargs)
    """
    timeout = timeout if timeout else None  # convert 0 or 0.0 to None
    # NB: if the loop is not running *yet*, it is OK to submit work
    # and we will wait for it
    if loop.is_closed():
        raise RuntimeError("Loop is not running")
    try:
        loop0 = asyncio.events.get_running_loop()
        if loop0 is loop:
            raise NotImplementedError("Calling sync() from within a running loop")  # noqa: TRY301
    except NotImplementedError:
        raise
    except RuntimeError:
        pass
    coro = func(*args, **kwargs)
    result: list[Any] = [None]
    event = threading.Event()
    asyncio.run_coroutine_threadsafe(_runner(event, coro, result, timeout), loop)
    # this timeout tracks wall-clock time for the threading.Event wait
    remaining_timeout = timeout
    while True:
        # this loop allows thread to get interrupted
        if event.wait(1):
            break
        if remaining_timeout is not None:
            remaining_timeout -= 1
            if remaining_timeout < 0:
                raise FSTimeoutError

    return_result = result[0]
    if isinstance(return_result, asyncio.TimeoutError):
        # suppress asyncio.TimeoutError, raise FSTimeoutError
        raise FSTimeoutError from return_result
    if isinstance(return_result, BaseException):
        raise return_result
    return return_result  # type: ignore[return-value]


def sync_wrapper[**P, T](func: Callable[P, Awaitable[T]], obj: Any = None) -> Callable[P, T]:
    """Given a function, make so can be called in blocking contexts.

    Leave obj=None if defining within a class. Pass the instance if attaching
    as an attribute of the instance.
    """

    @functools.wraps(func)
    def wrapper(*args: P.args, **kwargs: P.kwargs) -> T:
        self = obj or args[0]
        return sync(self.loop, func, *args, **kwargs)  # type: ignore[arg-type, union-attr]

    return wrapper  # type: ignore[return-value]


def wrap[**P, R](func: Callable[P, R]) -> Callable[P, Awaitable[R]]:
    @wraps(func)
    async def run(*args: P.args, **kwargs: P.kwargs) -> R:
        loop = get_running_loop()
        pfunc = partial(func, *args, **kwargs)
        return await loop.run_in_executor(None, pfunc)

    return run



"""Filesystem that automatically expands archive/structured files into virtual directories.

This filesystem wraps another filesystem and automatically treats certain file types
(like .zip, .tar.gz, .md) as directories that can be browsed into.

Example:
    ```python
    from fsspec import filesystem
    from upathtools.filesystems.auto_expand_fs import AutoExpandFS

    # Wrap a local filesystem
    local_fs = filesystem("file")
    auto_fs = AutoExpandFS(fs=local_fs)

    # Now .zip files appear as directories
    auto_fs.ls("/path/to/archive.zip")  # Lists contents of the zip
    auto_fs.cat("/path/to/archive.zip/inner/file.txt")  # Reads file inside zip

    # .md files can be browsed by headers
    auto_fs.ls("/path/to/readme.md")  # Lists markdown sections
    ```
"""

from __future__ import annotations

from typing import TYPE_CHECKING, Any, Literal, overload

import fsspec
from fsspec.asyn import AsyncFileSystem

from upathtools.async_ops import to_async_fs
from upathtools.filesystems.base import BaseAsyncFileSystem, BaseUPath, FileInfo


if TYPE_CHECKING:
    from collections.abc import Callable

    from fsspec.spec import AbstractFileSystem


class AutoExpandInfo(FileInfo, total=False):
    """Info dict for auto-expand filesystem paths."""

    size: int
    expanded: bool
    original_type: str


class AutoExpandPath(BaseUPath[AutoExpandInfo]):
    """UPath implementation for AutoExpandFS."""

    __slots__ = ()


# Default handlers mapping extension to (protocol, fs_kwargs_factory)
# The factory takes the parent fs and file path, returns kwargs for the expansion fs
type ExpanderFactory = Callable[[AsyncFileSystem, str], dict[str, Any]]

DEFAULT_EXPANDERS: dict[str, tuple[str, ExpanderFactory]] = {
    ".zip": ("zip", lambda fs, path: {"fo": path, "target_protocol": fs.protocol}),
    ".tar": ("tar", lambda fs, path: {"fo": path, "target_protocol": fs.protocol}),
    ".tar.gz": ("tar", lambda fs, path: {"fo": path, "target_protocol": fs.protocol}),
    ".tgz": ("tar", lambda fs, path: {"fo": path, "target_protocol": fs.protocol}),
    ".tar.bz2": ("tar", lambda fs, path: {"fo": path, "target_protocol": fs.protocol}),
    ".tar.xz": ("tar", lambda fs, path: {"fo": path, "target_protocol": fs.protocol}),
}


class AutoExpandFS(BaseAsyncFileSystem[AutoExpandPath, AutoExpandInfo]):
    """Filesystem that automatically expands archive/structured files as directories.

    When accessing a path like `/data/archive.zip/inner/file.txt`, this filesystem:
    1. Detects that `archive.zip` is an expandable file type
    2. Creates an appropriate sub-filesystem (e.g., ZipFileSystem)
    3. Delegates the remaining path (`inner/file.txt`) to that sub-filesystem

    Supports:
    - .zip files (via fsspec's zip protocol)
    - .tar, .tar.gz, .tgz, .tar.bz2, .tar.xz (via fsspec's tar protocol)
    - Custom expanders can be registered
    """

    protocol = "autoexpand"
    root_marker = "/"
    upath_cls = AutoExpandPath

    def __init__(
        self,
        fs: AbstractFileSystem | None = None,
        target_protocol: str | None = None,
        target_options: dict[str, Any] | None = None,
        expanders: dict[str, tuple[str, ExpanderFactory]] | None = None,
        **storage_options: Any,
    ) -> None:
        """Initialize auto-expand filesystem.

        Args:
            fs: Filesystem to wrap. If None, created from target_protocol.
            target_protocol: Protocol for target filesystem if fs not provided.
            target_options: Options for target filesystem if fs not provided.
            expanders: Custom expanders mapping extension to (protocol, kwargs_factory).
            **storage_options: Additional storage options.
        """
        super().__init__(**storage_options)

        if fs is None:
            if target_protocol is None:
                target_protocol = "file"
            fs = fsspec.filesystem(target_protocol, **(target_options or {}))
        assert fs
        # Wrap non-async filesystems
        self.fs = to_async_fs(fs)
        self.expanders = {**DEFAULT_EXPANDERS, **(expanders or {})}
        self._expansion_cache: dict[str, AsyncFileSystem] = {}

    def _get_expansion_ext(self, path: str) -> str | None:
        """Check if path has an expandable extension.

        Returns the matching extension or None.
        """
        path_lower = path.lower()
        # Check longer extensions first (e.g., .tar.gz before .gz)
        for ext in sorted(self.expanders.keys(), key=len, reverse=True):
            if path_lower.endswith(ext):
                return ext
        return None

    def _parse_path(self, path: str) -> tuple[AsyncFileSystem, str]:
        """Parse path and return (filesystem, remaining_path).

        If path contains an expandable file, returns the expansion filesystem
        and the path within that filesystem.
        """
        if not path or path == self.root_marker:
            return self.fs, self.root_marker

        # Normalize path
        path = path.lstrip("/")
        parts = path.split("/")

        # Walk through path parts looking for expandable files
        current_path = ""
        for i, part in enumerate(parts):
            current_path = f"{current_path}/{part}" if current_path else part

            ext = self._get_expansion_ext(current_path)
            if ext is not None:
                # Check if this is actually a file on the underlying fs
                try:
                    info = self.fs.info(current_path)
                    if info.get("type") == "file":
                        # Get or create expansion filesystem
                        expansion_fs = self._get_expansion_fs(current_path, ext)
                        # Remaining path is everything after this part
                        remaining = "/".join(parts[i + 1 :])
                        return expansion_fs, remaining or "/"
                except FileNotFoundError:
                    # Not a file, continue checking
                    pass

        # No expansion needed
        return self.fs, path

    def _get_expansion_fs(self, file_path: str, ext: str) -> AsyncFileSystem:
        """Get or create filesystem for expanding a file."""
        from upathtools import core

        if file_path in self._expansion_cache:
            return self._expansion_cache[file_path]

        protocol, kwargs_factory = self.expanders[ext]
        kwargs = kwargs_factory(self.fs, file_path)
        expansion_fs = core.filesystem(protocol, **kwargs)
        # Wrap non-async filesystems
        expansion_fs = to_async_fs(expansion_fs)

        self._expansion_cache[file_path] = expansion_fs
        return expansion_fs

    async def _info(self, path: str, **kwargs: Any) -> AutoExpandInfo:
        """Get info about a path."""
        fs, inner_path = self._parse_path(path)

        if fs is self.fs:
            # Check if this is an expandable file
            info = self.fs.info(inner_path)
            ext = self._get_expansion_ext(inner_path)
            if ext and info.get("type") == "file":
                # Expandable file appears as directory
                return AutoExpandInfo(
                    name=inner_path,
                    size=info.get("size", 0),
                    type="directory",
                    expanded=True,
                    original_type="file",
                )
            return AutoExpandInfo(
                name=info.get("name", inner_path),
                size=info.get("size", 0),
                type=info.get("type", "file"),
                expanded=False,
            )

        # Delegating to expansion filesystem
        info = await fs._info(inner_path, **kwargs)
        return AutoExpandInfo(
            name=info.get("name", inner_path),
            size=info.get("size", 0),
            type=info.get("type", "file"),
            expanded=True,
        )

    @overload
    async def _ls(
        self, path: str, detail: Literal[True] = ..., **kwargs: Any
    ) -> list[AutoExpandInfo]: ...

    @overload
    async def _ls(self, path: str, detail: bool, **kwargs: Any) -> list[str]: ...

    async def _ls(
        self, path: str, detail: bool = True, **kwargs: Any
    ) -> list[str] | list[AutoExpandInfo]:
        """List contents of a path."""
        fs, inner_path = self._parse_path(path)

        if fs is self.fs:
            # Check if path is an expandable file
            ext = self._get_expansion_ext(path.lstrip("/"))
            if ext:
                try:
                    info = self.fs.info(path.lstrip("/"))
                    if info.get("type") == "file":
                        # List contents of expanded file
                        expansion_fs = self._get_expansion_fs(path.lstrip("/"), ext)
                        items = await expansion_fs._ls("/", detail=True)

                        if not detail:
                            return [f"{path}/{item['name'].lstrip('/')}" for item in items]

                        return [
                            AutoExpandInfo(
                                name=f"{path}/{item['name'].lstrip('/')}",
                                size=item.get("size", 0),
                                type=item.get("type", "file"),
                                expanded=True,
                            )
                            for item in items
                        ]
                except FileNotFoundError:
                    pass

            # Regular directory listing
            items = self.fs.ls(inner_path, detail=True)
            results: list[AutoExpandInfo] = []

            for item in items:
                name = item.get("name", "")
                item_ext = self._get_expansion_ext(name)
                is_expandable = item_ext and item.get("type") == "file"

                results.append(
                    AutoExpandInfo(
                        name=name,
                        size=item.get("size", 0),
                        type="directory" if is_expandable else item.get("type", "file"),
                        expanded=bool(is_expandable),
                        original_type=item.get("type", "file") if is_expandable else None,  # type: ignore[typeddict-item]
                    )
                )

            if not detail:
                return [r["name"] for r in results]
            return results
        # Delegating to expansion filesystem
        items = await fs._ls(inner_path, detail=True)

        if not detail:
            return [item["name"] for item in items]

        return [
            AutoExpandInfo(
                name=item.get("name", ""),
                size=item.get("size", 0),
                type=item.get("type", "file"),
                expanded=True,
            )
            for item in items
        ]

    async def _cat_file(
        self, path: str, start: int | None = None, end: int | None = None, **kwargs: Any
    ) -> bytes:
        """Read file contents."""
        fs, inner_path = self._parse_path(path)
        return await fs._cat_file(inner_path, start=start, end=end, **kwargs)

    async def _pipe_file(self, path: str, value: bytes, **kwargs: Any) -> None:
        """Write file contents."""
        fs, inner_path = self._parse_path(path)

        # Don't allow writing into expanded archives
        if fs is not self.fs:
            msg = "Cannot write into expanded archives"
            raise NotImplementedError(msg)

        await fs._pipe_file(inner_path, value, **kwargs)

    async def _exists(self, path: str) -> bool:
        """Check if path exists."""
        fs, inner_path = self._parse_path(path)
        return await fs._exists(inner_path)

    async def _isfile(self, path: str) -> bool:
        """Check if path is a file."""
        info = await self._info(path)
        return info.get("type") == "file"

    async def _isdir(self, path: str) -> bool:
        """Check if path is a directory."""
        info = await self._info(path)
        return info.get("type") == "directory"

    def register_expander(
        self,
        extension: str,
        protocol: str,
        kwargs_factory: ExpanderFactory,
    ) -> None:
        """Register a custom expander for a file extension.

        Args:
            extension: File extension (e.g., ".md", ".json")
            protocol: fsspec protocol to use for expansion
            kwargs_factory: Factory function that takes (fs, path) and returns
                           kwargs for creating the expansion filesystem
        """
        self.expanders[extension] = (protocol, kwargs_factory)

    def clear_cache(self) -> None:
        """Clear the expansion filesystem cache."""
        self._expansion_cache.clear()


if __name__ == "__main__":
    from pathlib import Path
    import tempfile

    from upathtools import core

    # Create test files
    with tempfile.TemporaryDirectory() as tmpdir:
        # Create a markdown file
        md_content = """\
# Section 1
Content 1
## Section 1.1
Content 1.1
# Section 2
Content 2
"""
        md_path = Path(tmpdir) / "readme.md"
        md_path.write_text(md_content)

        # Create AutoExpandFS with markdown support
        local_fs = core.filesystem("file")
        auto_fs = AutoExpandFS(
            fs=local_fs,
            expanders={
                ".md": (
                    "md",
                    lambda fs, path: {"fo": path, "target_protocol": fs.protocol},
                ),
            },
        )

        # List directory - readme.md appears as directory
        print("Directory listing:")
        for item in auto_fs.ls(tmpdir, detail=True):
            print(f"  {item['name']} ({item['type']})")

        # List markdown sections
        print("\nMarkdown sections:")
        for item in auto_fs.ls(f"{tmpdir}/readme.md", detail=True):
            print(f"  {item['name']} ({item['type']})")



"""Combining filesystems that are composed of multiple filesystems."""

from __future__ import annotations

from .flat_union_fs import FlatUnionFileSystem, FlatUnionInfo, FlatUnionPath
from .union_fs import UnionFileSystem, UnionInfo, UnionPath
from .overlay_fs import OverlayFileSystem, OverlayInfo, OverlayPath

__all__ = [
    "FlatUnionFileSystem",
    "FlatUnionInfo",
    "FlatUnionPath",
    "OverlayFileSystem",
    "OverlayInfo",
    "OverlayPath",
    "UnionFileSystem",
    "UnionInfo",
    "UnionPath",
]



"""SFTP filesystem with UPath integration (requires paramiko)."""

from __future__ import annotations

from fsspec.implementations.sftp import SFTPFileSystem as _SFTPFileSystem
from upath.implementations.sftp import SFTPPath

from upathtools.filesystems.base import BaseFileSystem


class SFTPFileSystem(BaseFileSystem[SFTPPath], _SFTPFileSystem):
    """SFTP filesystem with UPath integration."""

    upath_cls = SFTPPath



"""E2B async filesystem implementation for upathtools."""

from __future__ import annotations

import base64
import logging
import os
from typing import TYPE_CHECKING, Any, Literal, Required, overload

from upathtools.async_helpers import sync_wrapper
from upathtools.filesystems.base import BaseAsyncFileSystem, BaseUPath, FileInfo


if TYPE_CHECKING:
    from upathtools.filesystems.base import CreationMode


logger = logging.getLogger(__name__)


class E2BInfo(FileInfo, total=False):
    """Info dict for E2B filesystem paths."""

    size: Required[int]
    mtime: Required[float]
    mode: Required[int]
    permissions: Required[str]
    owner: Required[str]
    group: Required[str]
    symlink_target: Required[str | None]


class E2BPath(BaseUPath[E2BInfo]):
    """E2B-specific UPath implementation."""

    __slots__ = ()


class E2BFS(BaseAsyncFileSystem[E2BPath, E2BInfo]):
    """Async filesystem for E2B sandbox environments.

    This filesystem provides access to files within an E2B sandbox environment,
    allowing you to read, write, and manipulate files remotely through the
    E2B native filesystem interface.
    """

    protocol = "e2b"
    upath_cls = E2BPath
    root_marker = "/"
    cachable = False  # Disable fsspec caching to prevent instance sharing

    def __init__(
        self,
        sandbox_id: str | None = None,
        api_key: str | None = None,
        template: str = "code-interpreter-v1",
        timeout: float = 300,
        **kwargs: Any,
    ) -> None:
        """Initialize E2B filesystem.

        Args:
            sandbox_id: Existing sandbox ID to connect to
            api_key: E2B API key
            template: E2B template to use for new sandboxes
            timeout: Default timeout for operations
            **kwargs: Additional filesystem options
        """
        super().__init__(**kwargs)
        self._sandbox_id = sandbox_id
        self._api_key = api_key
        self._template = template
        self._timeout = timeout
        self._sandbox = None
        self._session_started = False

    @staticmethod
    def _get_kwargs_from_urls(path: str) -> dict[str, Any]:
        path = path.removeprefix("e2b://")
        return {"sandbox_id": path}

    async def _get_sandbox(self):
        """Get or create E2B sandbox instance."""
        from e2b_code_interpreter import AsyncSandbox

        if self._sandbox is not None:
            return self._sandbox

        if self._sandbox_id:
            # Connect to existing sandbox
            self._sandbox = await AsyncSandbox.connect(self._sandbox_id, api_key=self._api_key)
        else:
            # Create new sandbox
            self._sandbox = await AsyncSandbox.create(self._template, api_key=self._api_key)
            assert self._sandbox
            self._sandbox_id = self._sandbox.sandbox_id

        return self._sandbox

    async def set_session(self) -> None:
        """Initialize the E2B session."""
        if not self._session_started:
            await self._get_sandbox()
            self._session_started = True

    async def close_session(self) -> None:
        """Close the E2B session."""
        if self._sandbox and self._session_started:
            await self._sandbox.kill()
            self._sandbox = None
            self._session_started = False

    @overload
    async def _ls(self, path: str, detail: Literal[True] = ..., **kwargs: Any) -> list[E2BInfo]: ...

    @overload
    async def _ls(self, path: str, detail: bool, **kwargs: Any) -> list[str]: ...

    async def _ls(self, path: str, detail: bool = True, **kwargs: Any) -> list[E2BInfo] | list[str]:
        """List directory contents with caching."""
        await self.set_session()
        sandbox = await self._get_sandbox()

        try:
            items = await sandbox.files.list(path)
        except Exception as exc:
            if "not found" in str(exc).lower() or "no such file" in str(exc).lower():
                raise FileNotFoundError(path) from exc
            msg = f"Failed to list directory {path}: {exc}"
            raise OSError(msg) from exc

        if not detail:
            return [item.path for item in items]

        result = []
        for item in items:
            try:
                info = await sandbox.files.get_info(item.path)
                from e2b.sandbox.filesystem.filesystem import FileType

                result.append(
                    E2BInfo(
                        name=item.path,
                        size=info.size,
                        type="directory" if info.type == FileType.DIR else "file",
                        mtime=info.modified_time.timestamp() if info.modified_time else 0,
                        mode=info.mode,
                        permissions=info.permissions,
                        owner=info.owner,
                        group=info.group,
                        symlink_target=info.symlink_target,
                    )
                )
            except Exception:  # noqa: BLE001
                # Fallback if get_info fails - use item.is_dir
                result.append(
                    E2BInfo(
                        name=item.path,
                        size=0,
                        type="directory" if item.type == FileType.DIR else "file",
                        mtime=0,
                        mode=0,
                        permissions="",
                        owner="",
                        group="",
                        symlink_target=None,
                    )
                )

        return result

    async def _cat_file(
        self, path: str, start: int | None = None, end: int | None = None, **kwargs: Any
    ) -> bytes:
        """Read file contents."""
        await self.set_session()
        sandbox = await self._get_sandbox()

        try:
            content = await sandbox.files.read(path)
        except Exception as exc:
            if "not found" in str(exc).lower() or "no such file" in str(exc).lower():
                raise FileNotFoundError(path) from exc
            if "is a directory" in str(exc).lower():
                raise IsADirectoryError(path) from exc
            msg = f"Failed to read file {path}: {exc}"
            raise OSError(msg) from exc

        # Handle byte ranges if specified
        if isinstance(content, str):
            content = content.encode("utf-8")

        if start is not None or end is not None:
            start = start or 0
            end = end or len(content)
            content = content[start:end]

        return content

    async def _put_file(
        self,
        lpath: str,
        rpath: str,
        callback=None,
        **kwargs: Any,
    ) -> None:
        """Upload a local file to the sandbox."""
        await self.set_session()

        # Read local file
        with open(lpath, "rb") as f:  # noqa: PTH123
            data = f.read()

        await self._pipe_file(rpath, data, **kwargs)

    async def _pipe_file(
        self, path: str, value: bytes, mode: CreationMode = "overwrite", **kwargs: Any
    ) -> None:
        """Write data to a file in the sandbox."""
        await self.set_session()
        sandbox = await self._get_sandbox()

        try:
            # Convert bytes to string for E2B - they handle encoding
            if isinstance(value, bytes):
                try:
                    content = value.decode("utf-8")
                except UnicodeDecodeError:
                    # For binary files, encode as base64 and write a script to decode
                    encoded = base64.b64encode(value).decode("ascii")
                    decode_script = f"""
import base64
with open({path!r}, 'wb') as f:
    f.write(base64.b64decode({encoded!r}))
"""
                    # Execute the decode script
                    execution = await sandbox.run_code(decode_script)
                    if execution.error:
                        msg = f"Failed to write binary file: {execution.error.value}"
                        raise OSError(msg)  # noqa: B904
                    return
            else:
                content = value

            await sandbox.files.write(path, content)
        except Exception as exc:
            msg = f"Failed to write file {path}: {exc}"
            raise OSError(msg) from exc

    async def _mkdir(self, path: str, create_parents: bool = True, **kwargs: Any) -> None:
        """Create a directory."""
        await self.set_session()
        sandbox = await self._get_sandbox()

        try:
            await sandbox.files.make_dir(path)
        except Exception as exc:
            # E2B make_dir might not have create_parents option
            if create_parents and "parent" in str(exc).lower():
                # Try to create parent directories first
                parent = os.path.dirname(path)  # noqa: PTH120
                if parent and parent not in (path, "/"):
                    await self._mkdir(parent, create_parents=True)
                    await sandbox.files.make_dir(path)
            else:
                msg = f"Failed to create directory {path}: {exc}"
                raise OSError(msg) from exc

    async def _rm_file(self, path: str, **kwargs: Any) -> None:
        """Remove a file."""
        await self.set_session()
        sandbox = await self._get_sandbox()

        try:
            await sandbox.files.remove(path)
        except Exception as exc:
            if "not found" in str(exc).lower() or "no such file" in str(exc).lower():
                raise FileNotFoundError(path) from exc
            if "is a directory" in str(exc).lower():
                raise IsADirectoryError(path) from exc
            msg = f"Failed to remove file {path}: {exc}"
            raise OSError(msg) from exc

    async def _rmdir(self, path: str, **kwargs: Any) -> None:
        """Remove a directory."""
        await self.set_session()
        sandbox = await self._get_sandbox()

        try:
            await sandbox.files.remove(path)
        except Exception as exc:
            if "not found" in str(exc).lower() or "no such file" in str(exc).lower():
                raise FileNotFoundError(path) from exc
            if "not a directory" in str(exc).lower():
                raise NotADirectoryError(path) from exc
            if "not empty" in str(exc).lower():
                msg = f"Directory not empty: {path}"
                raise OSError(msg) from exc
            msg = f"Failed to remove directory {path}: {exc}"
            raise OSError(msg) from exc

    async def _exists(self, path: str, **kwargs: Any) -> bool:
        """Check if path exists."""
        await self.set_session()
        sandbox = await self._get_sandbox()

        try:
            return await sandbox.files.exists(path)
        except Exception:  # noqa: BLE001
            return False

    async def _isfile(self, path: str, **kwargs: Any) -> bool:
        """Check if path is a file."""
        from e2b.sandbox.filesystem.filesystem import FileType

        await self.set_session()
        sandbox = await self._get_sandbox()
        try:
            if not await sandbox.files.exists(path):
                return False
            info = await sandbox.files.get_info(path)
            return info.type == FileType.FILE  # noqa: TRY300
        except Exception:  # noqa: BLE001
            return False

    async def _isdir(self, path: str, **kwargs: Any) -> bool:
        """Check if path is a directory."""
        from e2b.sandbox.filesystem.filesystem import FileType

        await self.set_session()
        sandbox = await self._get_sandbox()
        try:
            if not await sandbox.files.exists(path):
                return False
            info = await sandbox.files.get_info(path)
            return info.type == FileType.DIR  # noqa: TRY300
        except Exception:  # noqa: BLE001
            return False

    async def _size(self, path: str, **kwargs: Any) -> int:
        """Get file size."""
        await self.set_session()
        sandbox = await self._get_sandbox()

        try:
            info = await sandbox.files.get_info(path)
        except Exception as exc:
            if "not found" in str(exc).lower() or "no such file" in str(exc).lower():
                raise FileNotFoundError(path) from exc
            raise OSError(f"Failed to get file size for {path}: {exc}") from exc
        else:
            return info.size

    async def _modified(self, path: str, **kwargs: Any) -> float:
        """Get file modification time."""
        await self.set_session()
        sandbox = await self._get_sandbox()

        try:
            info = await sandbox.files.get_info(path)
            return info.modified_time.timestamp() if info.modified_time else 0.0
        except Exception as exc:
            if "not found" in str(exc).lower() or "no such file" in str(exc).lower():
                raise FileNotFoundError(path) from exc
            raise OSError(f"Failed to get modification time for {path}: {exc}") from exc

    async def _info(self, path: str, **kwargs: Any) -> E2BInfo:
        """Get info about a file or directory."""
        from e2b.sandbox.filesystem.filesystem import FileType

        await self.set_session()
        sandbox = await self._get_sandbox()
        try:
            info = await sandbox.files.get_info(path)
            return E2BInfo(
                name=path,
                size=info.size,
                type="directory" if info.type == FileType.DIR else "file",
                mtime=info.modified_time.timestamp() if info.modified_time else 0.0,
                mode=info.mode,
                permissions=info.permissions,
                owner=info.owner,
                group=info.group,
                symlink_target=info.symlink_target,
            )
        except Exception as exc:
            if "not found" in str(exc).lower() or "no such file" in str(exc).lower():
                raise FileNotFoundError(path) from exc
            msg = f"Failed to get info for {path}: {exc}"
            raise OSError(msg) from exc

    @overload
    async def _find(
        self,
        path: str,
        maxdepth: int | None = None,
        withdirs: bool = False,
        *,
        detail: Literal[False] = False,
        **kwargs: Any,
    ) -> list[str]: ...

    @overload
    async def _find(
        self,
        path: str,
        maxdepth: int | None = None,
        withdirs: bool = False,
        *,
        detail: Literal[True],
        **kwargs: Any,
    ) -> dict[str, E2BInfo]: ...

    async def _find(
        self,
        path: str,
        maxdepth: int | None = None,
        withdirs: bool = False,
        *,
        detail: bool = False,
        **kwargs: Any,
    ) -> list[str] | dict[str, E2BInfo]:
        """Recursively list all files using E2B's depth parameter.

        More efficient than walking the directory tree with multiple _ls calls.
        """
        from e2b.sandbox.filesystem.filesystem import FileType

        await self.set_session()
        sandbox = await self._get_sandbox()
        # Use a large depth for recursive listing (E2B doesn't support None for unlimited)
        depth = maxdepth if maxdepth is not None else 100
        try:
            items = await sandbox.files.list(path, depth=depth)
        except Exception as exc:
            if "not found" in str(exc).lower() or "no such file" in str(exc).lower():
                raise FileNotFoundError(path) from exc
            msg = f"Failed to find files in {path}: {exc}"
            raise OSError(msg) from exc
        else:
            filtered = items if withdirs else [item for item in items if item.type == FileType.FILE]
            if detail:
                return {
                    item.path: E2BInfo(
                        name=item.path,
                        type="directory" if item.type == FileType.DIR else "file",
                        size=0,  # E2B list doesn't return size
                        mtime=0.0,
                        mode=0,
                        permissions="",
                        owner="",
                        group="",
                        symlink_target=None,
                    )
                    for item in filtered
                }
            return [item.path for item in filtered]

    # TODO: Add _glob and _grep using sandbox.commands.run() with Linux CLI tools (find, grep)
    # This would be faster than fsspec's default implementations.

    # Sync wrappers for async methods
    ls = sync_wrapper(_ls)  # pyright: ignore[reportAssignmentType]
    cat_file = sync_wrapper(_cat_file)  # pyright: ignore[reportAssignmentType]
    pipe_file = sync_wrapper(_pipe_file)  # pyright: ignore[reportAssignmentType]
    mkdir = sync_wrapper(_mkdir)
    rm_file = sync_wrapper(_rm_file)
    rmdir = sync_wrapper(_rmdir)
    exists = sync_wrapper(_exists)  # pyright: ignore[reportAssignmentType]
    isfile = sync_wrapper(_isfile)
    isdir = sync_wrapper(_isdir)
    size = sync_wrapper(_size)
    modified = sync_wrapper(_modified)
    info = sync_wrapper(_info)
    find = sync_wrapper(_find)  # pyright: ignore[reportAssignmentType]



"""Filesystem implementation for browsing Pydantic BaseModel instance data."""

from __future__ import annotations

import json
from typing import TYPE_CHECKING, Any, Literal, overload

from upathtools.filesystems.base import BaseAsyncFileSystem, BaseUPath, FileInfo


if TYPE_CHECKING:
    from pydantic import BaseModel


class BaseModelInstanceInfo(FileInfo, total=False):
    """Info dict for BaseModel instance paths."""

    entry_kind: Literal["instance", "nested_object", "value", "special", "key", "item", "field"]
    class_name: str
    is_basemodel: bool
    field_count: int
    data: str
    is_collection: bool
    size: int | None
    value_type: str
    value: str
    description: str
    is_nested: bool
    index: int


class BaseModelInstancePath(BaseUPath[BaseModelInstanceInfo]):
    """UPath implementation for browsing Pydantic BaseModel instance data."""

    __slots__ = ()

    def iterdir(self):
        if not self.is_dir():
            raise NotADirectoryError(str(self))
        yield from super().iterdir()

    @property
    def path(self) -> str:
        path = super().path
        return "/" if path == "." else path


class BaseModelInstanceFileSystem(
    BaseAsyncFileSystem[BaseModelInstancePath, BaseModelInstanceInfo]
):
    """Filesystem for browsing Pydantic BaseModel instance data and values."""

    protocol = "basemodel-instance"
    upath_cls = BaseModelInstancePath

    def __init__(self, instance: BaseModel, **kwargs: Any) -> None:
        """Initialize the filesystem.

        Args:
            instance: BaseModel instance to browse
            kwargs: Additional keyword arguments for the filesystem
        """
        super().__init__(**kwargs)
        self.instance = instance

    @overload
    async def _ls(
        self,
        path: str,
        detail: Literal[True] = ...,
        **kwargs: Any,
    ) -> list[BaseModelInstanceInfo]: ...

    @overload
    async def _ls(self, path: str, detail: bool, **kwargs: Any) -> list[str]: ...

    async def _ls(
        self,
        path: str,
        detail: bool = True,
        **kwargs: Any,
    ) -> list[BaseModelInstanceInfo] | list[str]:
        """List instance fields and values."""
        path = self._strip_protocol(path).strip("/")  # pyright: ignore[reportAttributeAccessIssue]

        try:
            current_obj, field_name = _get_nested_value_at_path(self.instance, path)
        except FileNotFoundError:
            return []

        if field_name:
            # Listing a specific field value
            if not hasattr(current_obj, field_name):
                return []

            field_value = getattr(current_obj, field_name)

            if _is_basemodel_instance(field_value):
                # BaseModel instance - show its fields
                items = list(type(field_value).model_fields.keys())
                items.extend(["__json__", "__dict__", "__schema__"])
            elif _is_list_like(field_value):
                # List-like - show indices
                items = [str(i) for i in range(len(field_value))]
                items.extend(["__json__", "__length__", "__type__"])
            elif _is_dict_like(field_value):
                # Dict-like - show keys
                items = list(field_value.keys())
                items.extend(["__json__", "__keys__", "__values__"])
            else:
                # Primitive value - show special paths
                items = ["__value__", "__type__", "__str__", "__repr__"]
        # Listing model root - show all fields plus special paths
        elif _is_basemodel_instance(current_obj):
            items = list(type(current_obj).model_fields.keys())
            items.extend(["__json__", "__dict__", "__schema__", "__model_dump__"])
        elif _is_list_like(current_obj):
            items = [str(i) for i in range(len(current_obj))]
            items.extend(["__json__", "__length__", "__type__"])
        elif _is_dict_like(current_obj):
            items = list(current_obj.keys())
            items.extend(["__json__", "__keys__", "__values__"])
        else:
            items = ["__value__", "__type__", "__str__", "__repr__"]

        if not detail:
            return items

        result = []
        for item in items:
            if item.startswith("__"):
                desc = f"Special path for {item[2:-2]} information"
                info = BaseModelInstanceInfo(
                    name=item, type="file", entry_kind="special", size=0, description=desc
                )
                result.append(info)
            # It's a field or item
            elif _is_basemodel_instance(current_obj):
                field_value = getattr(current_obj, item)
                is_navigable = (
                    _is_basemodel_instance(field_value)
                    or _is_list_like(field_value)
                    or _is_dict_like(field_value)
                )
                result.append(
                    BaseModelInstanceInfo(
                        name=item,
                        type="directory" if is_navigable else "file",
                        entry_kind="field",
                        value_type=type(field_value).__name__,
                        value=str(field_value)[:100] + "..."
                        if len(str(field_value)) > 100  # noqa: PLR2004
                        else str(field_value),
                        is_nested=_is_basemodel_instance(field_value),
                        is_collection=_is_list_like(field_value) or _is_dict_like(field_value),
                    )
                )
            elif _is_list_like(current_obj):
                idx = int(item)
                item_value = current_obj[idx]
                is_navigable = (
                    _is_basemodel_instance(item_value)
                    or _is_list_like(item_value)
                    or _is_dict_like(item_value)
                )
                result.append(
                    BaseModelInstanceInfo(
                        name=item,
                        type="directory" if is_navigable else "file",
                        entry_kind="item",
                        index=idx,
                        value_type=type(item_value).__name__,
                        value=str(item_value)[:100] + "..."
                        if len(str(item_value)) > 100  # noqa: PLR2004
                        else str(item_value),
                        is_nested=_is_basemodel_instance(item_value),
                    )
                )
            elif _is_dict_like(current_obj):
                dict_value = current_obj[item]
                is_navigable = (
                    _is_basemodel_instance(dict_value)
                    or _is_list_like(dict_value)
                    or _is_dict_like(dict_value)
                )
                result.append(
                    BaseModelInstanceInfo(
                        name=item,
                        type="directory" if is_navigable else "file",
                        entry_kind="key",
                        value_type=type(dict_value).__name__,
                        value=str(dict_value)[:100] + "..."
                        if len(str(dict_value)) > 100  # noqa: PLR2004
                        else str(dict_value),
                        is_nested=_is_basemodel_instance(dict_value),
                    )
                )

        return result

    async def _isdir(self, path: str) -> bool:
        """Check if path is a directory (BaseModel instance, list, or dict)."""
        path = self._strip_protocol(path).strip("/")  # pyright: ignore[reportAttributeAccessIssue]

        if not path:
            # Root is always a directory
            return True

        try:
            current_obj, field_name = _get_nested_value_at_path(self.instance, path)
        except FileNotFoundError:
            return False

        if field_name:
            # Check if the field value is navigable
            if not hasattr(current_obj, field_name):
                return False
            field_value = getattr(current_obj, field_name)
            return (
                _is_basemodel_instance(field_value)
                or _is_list_like(field_value)
                or _is_dict_like(field_value)
            )

        # current_obj itself - check if it's navigable
        return (
            _is_basemodel_instance(current_obj)
            or _is_list_like(current_obj)
            or _is_dict_like(current_obj)
        )

    async def _cat_file(
        self, path: str, start: int | None = None, end: int | None = None, **kwargs: Any
    ) -> bytes:
        """Get field values, JSON representation, or other information."""
        data = self._get_content(path)
        return data[start:end]

    def _get_content(self, path: str) -> bytes:  # noqa: PLR0911
        """Get the raw content bytes for a path."""
        path = self._strip_protocol(path).strip("/")  # pyright: ignore[reportAttributeAccessIssue]

        if not path:
            # Return instance JSON
            if _is_basemodel_instance(self.instance):
                return self.instance.model_dump_json(indent=2).encode()
            return json.dumps(self.instance, indent=2, default=str).encode()

        parts = path.split("/")

        # Handle special paths
        if parts[-1].startswith("__") and parts[-1].endswith("__"):
            special_path = parts[-1]
            field_path = "/".join(parts[:-1])

            try:
                current_obj, field_name = _get_nested_value_at_path(self.instance, field_path)
            except FileNotFoundError:
                raise FileNotFoundError(f"Path {field_path} not found") from None

            target_obj = getattr(current_obj, field_name) if field_name else current_obj

            match special_path:
                case "__json__":
                    if _is_basemodel_instance(target_obj):
                        return target_obj.model_dump_json(indent=2).encode()
                    return json.dumps(target_obj, indent=2, default=str).encode()

                case "__dict__":
                    if _is_basemodel_instance(target_obj):
                        return json.dumps(target_obj.model_dump(), indent=2).encode()
                    if hasattr(target_obj, "__dict__"):
                        return json.dumps(target_obj.__dict__, indent=2, default=str).encode()
                    return json.dumps(dict(target_obj), indent=2, default=str).encode()

                case "__schema__":
                    if _is_basemodel_instance(target_obj):
                        return json.dumps(target_obj.model_json_schema(), indent=2).encode()

                    raise FileNotFoundError("Schema only available for BaseModel instances")

                case "__model_dump__":
                    if _is_basemodel_instance(target_obj):
                        return json.dumps(target_obj.model_dump(), indent=2).encode()

                    raise FileNotFoundError("model_dump only available for BaseModel instances")

                case "__value__":
                    return str(target_obj).encode()

                case "__type__":
                    return str(type(target_obj)).encode()

                case "__str__":
                    return str(target_obj).encode()

                case "__repr__":
                    return repr(target_obj).encode()

                case "__length__":
                    if hasattr(target_obj, "__len__"):
                        return str(len(target_obj)).encode()

                    raise FileNotFoundError(f"Length not available for {type(target_obj)}")

                case "__keys__":
                    if _is_dict_like(target_obj):
                        return json.dumps(list(target_obj.keys()), indent=2).encode()

                    raise FileNotFoundError("Keys only available for dict-like objects")

                case "__values__":
                    if _is_dict_like(target_obj):
                        return json.dumps(list(target_obj.values()), indent=2, default=str).encode()

                    raise FileNotFoundError("Values only available for dict-like objects")

                case _:
                    raise FileNotFoundError(f"Unknown special path: {special_path}")

        # Regular field/item path
        try:
            current_obj, field_name = _get_nested_value_at_path(self.instance, path)
        except FileNotFoundError:
            raise FileNotFoundError(f"Path {path} not found") from None

        if not field_name:
            # Return the object itself
            if _is_basemodel_instance(current_obj):
                return current_obj.model_dump_json(indent=2).encode()
            return json.dumps(current_obj, indent=2, default=str).encode()

        # Get the field value
        if _is_basemodel_instance(current_obj):
            if not hasattr(current_obj, field_name):
                raise FileNotFoundError(f"Field {field_name} not found")
            field_value = getattr(current_obj, field_name)
        elif _is_list_like(current_obj):
            try:
                idx = int(field_name)
                field_value = current_obj[idx]
            except (ValueError, IndexError) as exc:
                raise FileNotFoundError(f"Invalid index {field_name}") from exc
        elif _is_dict_like(current_obj):
            if field_name not in current_obj:
                raise FileNotFoundError(f"Key {field_name} not found")
            field_value = current_obj[field_name]
        else:
            raise FileNotFoundError(f"Cannot access {field_name} on {type(current_obj)}")

        # Return the field value
        if _is_basemodel_instance(field_value):
            return field_value.model_dump_json(indent=2).encode()
        return json.dumps(field_value, indent=2, default=str).encode()

    async def _info(self, path: str, **kwargs: Any) -> BaseModelInstanceInfo:
        """Get detailed info about an instance field or value."""
        path = self._strip_protocol(path).strip("/")  # pyright: ignore[reportAttributeAccessIssue]

        if not path:
            # Root instance info
            return BaseModelInstanceInfo(
                name=type(self.instance).__name__,
                type="directory",
                entry_kind="instance",
                class_name=f"{type(self.instance).__module__}.{type(self.instance).__name__}",
                is_basemodel=_is_basemodel_instance(self.instance),
                field_count=len(type(self.instance).model_fields)
                if _is_basemodel_instance(self.instance)
                else 0,
                data=str(self.instance)[:200] + "..."
                if len(str(self.instance)) > 200  # noqa: PLR2004
                else str(self.instance),
            )

        try:
            current_obj, field_name = _get_nested_value_at_path(self.instance, path)
        except FileNotFoundError as exc:
            raise FileNotFoundError(f"Path {path} not found") from exc

        if not field_name:
            # Nested object info
            is_navigable = (
                _is_basemodel_instance(current_obj)
                or _is_list_like(current_obj)
                or _is_dict_like(current_obj)
            )
            return BaseModelInstanceInfo(
                name=type(current_obj).__name__,
                type="directory" if is_navigable else "file",
                entry_kind="nested_object",
                class_name=f"{type(current_obj).__module__}.{type(current_obj).__name__}",
                is_basemodel=_is_basemodel_instance(current_obj),
                is_collection=_is_list_like(current_obj) or _is_dict_like(current_obj),
                size=len(current_obj) if hasattr(current_obj, "__len__") else None,
                data=str(current_obj)[:200] + "..."
                if len(str(current_obj)) > 200  # noqa: PLR2004
                else str(current_obj),
            )

        # Field/item value info
        if _is_basemodel_instance(current_obj):
            if not hasattr(current_obj, field_name):
                raise FileNotFoundError(f"Field {field_name} not found")
            field_value = getattr(current_obj, field_name)
        elif _is_list_like(current_obj):
            try:
                idx = int(field_name)
                field_value = current_obj[idx]
            except (ValueError, IndexError) as exc:
                raise FileNotFoundError(f"Invalid index {field_name}") from exc
        elif _is_dict_like(current_obj):
            if field_name not in current_obj:
                raise FileNotFoundError(f"Key {field_name} not found")
            field_value = current_obj[field_name]
        else:
            raise FileNotFoundError(f"Cannot access {field_name} on {type(current_obj)}")

        is_navigable = (
            _is_basemodel_instance(field_value)
            or _is_list_like(field_value)
            or _is_dict_like(field_value)
        )
        return BaseModelInstanceInfo(
            name=field_name,
            type="directory" if is_navigable else "file",
            entry_kind="value",
            value_type=type(field_value).__name__,
            value=str(field_value)[:200] + "..."
            if len(str(field_value)) > 200  # noqa: PLR2004
            else str(field_value),
            is_basemodel=_is_basemodel_instance(field_value),
            is_collection=_is_list_like(field_value) or _is_dict_like(field_value),
            size=len(field_value) if hasattr(field_value, "__len__") else None,
        )


def _is_basemodel_instance(obj: Any) -> bool:
    """Check if object is a BaseModel instance."""
    return hasattr(type(obj), "model_fields") and hasattr(obj, "model_dump")


def _is_list_like(obj: Any) -> bool:
    """Check if object is list-like."""
    return isinstance(obj, (list, tuple, set))


def _is_dict_like(obj: Any) -> bool:
    """Check if object is dict-like."""
    return isinstance(obj, dict)


def _get_nested_value_at_path(instance: BaseModel, path: str) -> tuple[Any, str]:
    """Get the object and field name at a given path."""
    if not path:
        return instance, ""

    parts = path.strip("/").split("/")
    current_obj = instance

    for i, part in enumerate(parts[:-1]):
        if part.startswith("__") and part.endswith("__"):
            # Skip special paths like __json__, __dict__
            continue

        if not hasattr(current_obj, part):
            raise FileNotFoundError(f"Field {part} not found in {type(current_obj).__name__}")

        current_obj = getattr(current_obj, part)

        # Handle list/dict access with numeric indices
        if isinstance(current_obj, (list, tuple)) and i + 1 < len(parts) - 1:
            next_part = parts[i + 1]
            if next_part.isdigit():
                idx = int(next_part)
                if idx >= len(current_obj):
                    raise FileNotFoundError(f"Index {idx} out of range for {part}")
                current_obj = current_obj[idx]
                parts.pop(i + 1)  # Remove the index from parts

    return current_obj, parts[-1] if parts else ""


if __name__ == "__main__":
    import asyncio

    from pydantic import BaseModel, Field

    class Address(BaseModel):
        street: str
        city: str
        country: str = "USA"

    class User(BaseModel):
        name: str = Field(min_length=1, max_length=50)
        age: int = Field(ge=0, le=120)
        email: str
        address: Address
        tags: list[str] = []

    async def main() -> None:
        user = User(
            name="John Doe",
            age=30,
            email="john@example.com",
            address=Address(street="123 Main St", city="New York"),
            tags=["developer", "python"],
        )

        fs = BaseModelInstanceFileSystem(user)
        print("Fields:", fs.ls("/", detail=False))
        print("User info:", fs.info("/"))
        print("Address info:", fs.info("/address"))
        print("Tags:", fs.cat_file("/tags"))

    asyncio.run(main())



"""GitHub Gist filesystem implementation with async support using httpx."""

from __future__ import annotations

import base64
import contextlib
import io
import logging
import os
from typing import TYPE_CHECKING, Any, Literal, overload
import weakref

from fsspec.utils import infer_storage_options

from upathtools.async_helpers import sync, sync_wrapper
from upathtools.filesystems.base import BaseAsyncFileSystem, BaseUPath, BufferedWriter, FileInfo


if TYPE_CHECKING:
    import httpx


class GistInfo(FileInfo, total=False):
    """Info dict for Gist filesystem paths."""

    size: int
    description: str | None
    created_at: str | None
    updated_at: str | None
    html_url: str | None
    git_pull_url: str | None
    git_push_url: str | None
    gist_id: str
    mime_type: str
    comments: int | None
    public: bool | None
    owner: str | None
    files_count: int | None
    files_preview: list[str] | None
    # File-specific fields from GitHub API
    filename: str | None
    raw_url: str | None
    language: str | None
    content: str | None
    truncated: bool | None


logger = logging.getLogger(__name__)


class GistPath(BaseUPath[GistInfo]):
    """UPath implementation for GitHub Gist filesystem."""

    __slots__ = ()


class GistFileSystem(BaseAsyncFileSystem[GistPath, GistInfo]):
    """Filesystem for accessing GitHub Gists files.

    Supports both individual gists and listing all gists for a user.
    Uses httpx for both synchronous and asynchronous operations.
    """

    protocol = "gist"
    upath_cls = GistPath
    gist_url = "https://api.github.com/gists/{gist_id}"
    gist_rev_url = "https://api.github.com/gists/{gist_id}/{sha}"
    user_gists_url = "https://api.github.com/users/{username}/gists"
    auth_gists_url = "https://api.github.com/gists"

    def __init__(
        self,
        gist_id: str | None = None,
        username: str | None = None,
        token: str | None = None,
        sha: str | None = None,
        timeout: int | None = None,
        asynchronous: bool = False,
        loop: Any = None,
        client_kwargs: dict[str, Any] | None = None,
        **kwargs: Any,
    ) -> None:
        """Initialize the filesystem.

        Args:
            gist_id: Specific gist ID to access
            username: GitHub username for listing all gists
            token: GitHub personal access token for authentication
            sha: Specific revision of a gist
            timeout: Connection timeout in seconds
            asynchronous: Whether to use async operations
            loop: Event loop for async operations
            client_kwargs: Additional arguments for httpx client
            **kwargs: Additional filesystem options
        """
        super().__init__(asynchronous=asynchronous, loop=loop, **kwargs)

        self.gist_id = gist_id
        self.username = username
        self.token = token or os.environ.get("GITHUB_TOKEN") or os.environ.get("GH_TOKEN")
        self.sha = sha
        self.timeout = timeout if timeout is not None else 60.0
        self.client_kwargs = client_kwargs or {}
        self._session: httpx.AsyncClient | None = None

        # We can work in two modes:
        # 1. Single gist mode (gist_id is provided)
        # 2. User gists mode (username is provided, or token alone for authenticated user)
        if not gist_id and not username and not self.token:
            msg = "Either gist_id, username, or token must be provided"
            raise ValueError(msg)

        self.headers = {"Authorization": f"token {self.token}"} if self.token else {}
        self.dircache: dict[str, Any] = {}

    @property
    def fsid(self) -> str:
        """Filesystem ID."""
        return "gist"

    async def set_session(self) -> httpx.AsyncClient:
        """Set up and return the httpx async client."""
        if self._session is None:
            import httpx

            self._session = httpx.AsyncClient(
                follow_redirects=True,
                timeout=self.timeout,
                headers=self.headers,
                **self.client_kwargs,
            )

            if not self.asynchronous:
                weakref.finalize(self, self.close_session, self.loop, self._session)

        return self._session

    @staticmethod
    def close_session(loop: Any, session: httpx.AsyncClient) -> None:
        """Close the httpx session."""
        if loop is not None and loop.is_running():
            with contextlib.suppress(TimeoutError, RuntimeError):
                sync(loop, session.aclose, timeout=0.1)

    @overload
    @classmethod
    def _strip_protocol(cls, path: str) -> str: ...

    @overload
    @classmethod
    def _strip_protocol(cls, path: list[str]) -> list[str]: ...

    @classmethod
    def _strip_protocol(cls, path: str | list[str]) -> str | list[str]:
        """Strip protocol prefix from path."""
        if isinstance(path, list):
            return [cls._strip_protocol(p) for p in path]
        p = infer_storage_options(path).get("path", path)
        return p.lstrip("/")

    @classmethod
    def _get_kwargs_from_urls(cls, path: str) -> dict[str, Any]:
        """Parse URL into constructor kwargs."""
        so = infer_storage_options(path)
        out = {}

        if so.get("username"):
            out["username"] = so["username"]
        if so.get("password"):
            out["token"] = so["password"]
        if so.get("host"):
            # The host could be a gist ID or a username
            host = so["host"]
            # Simple heuristic: gist IDs are typically 32 hex chars
            if len(host) == 32 and all(c in "0123456789abcdef" for c in host.lower()):  # noqa: PLR2004
                out["gist_id"] = host
            else:
                out["username"] = host

        return out

    async def _fetch_gist_metadata(self, gist_id: str) -> dict[str, Any]:
        """Fetch metadata for a specific gist.

        Args:
            gist_id: ID of the gist to fetch

        Returns:
            Dictionary containing gist metadata

        Raises:
            FileNotFoundError: If gist is not found
        """
        session = await self.set_session()

        if self.sha:
            url = self.gist_rev_url.format(gist_id=gist_id, sha=self.sha)
        else:
            url = self.gist_url.format(gist_id=gist_id)

        logger.debug("Fetching gist metadata: %s", url)
        response = await session.get(url)
        if response.status_code == 404:  # noqa: PLR2004
            msg = f"Gist not found: {gist_id}@{self.sha or 'latest'}"
            raise FileNotFoundError(msg)

        response.raise_for_status()
        return response.json()

    async def _fetch_user_gists(self, page: int = 1, per_page: int = 100) -> list[dict[str, Any]]:
        """Fetch gists for a user.

        Args:
            page: Page number for pagination
            per_page: Number of gists per page

        Returns:
            List of gist metadata dictionaries

        Raises:
            FileNotFoundError: If user is not found
        """
        session = await self.set_session()

        params = {"page": page, "per_page": per_page}
        if self.username and not self.token:
            url = self.user_gists_url.format(username=self.username)
        else:
            url = self.auth_gists_url

        logger.debug("Fetching user gists: %s", url)
        response = await session.get(url, params=params)

        if response.status_code == 404:  # noqa: PLR2004
            msg = f"User not found: {self.username}"
            raise FileNotFoundError(msg)

        response.raise_for_status()
        return response.json()

    async def _get_gist_file_list(self, gist_id: str) -> list[GistInfo]:
        """Get list of files in a specific gist.

        Args:
            gist_id: ID of the gist

        Returns:
            List of detailed file metadata dictionaries
        """
        if gist_id in self.dircache:
            return self.dircache[gist_id]

        # Fetch the specific gist metadata
        meta = await self._fetch_gist_metadata(gist_id)
        out = [
            GistInfo(
                name=fname,
                type="file",
                size=finfo.get("size", 0),
                raw_url=finfo.get("raw_url"),
                language=finfo.get("language"),
                mime_type=finfo.get("type", "text/plain"),
                gist_id=gist_id,
                description=meta.get("description", ""),
                html_url=meta.get("html_url"),
                created_at=meta.get("created_at"),
                updated_at=meta.get("updated_at"),
                public=meta.get("public", False),
                truncated=finfo.get("truncated", False),
            )
            for fname, finfo in meta.get("files", {}).items()
            if finfo is not None
        ]
        self.dircache[gist_id] = out
        return out

    async def _get_all_gists(self) -> list[GistInfo]:
        """Get metadata for all gists of the user."""
        if "" in self.dircache:
            return self.dircache[""]

        gists = await self._fetch_user_gists(page=1, per_page=100)
        all_gists = gists.copy()
        page = 2
        while len(gists) == 100:  # noqa: PLR2004
            gists = await self._fetch_user_gists(page=page, per_page=100)
            all_gists.extend(gists)
            page += 1
        out = [
            GistInfo(
                gist_id=gist["id"],
                name=gist["id"],
                type="directory",
                description=gist.get("description", ""),
                created_at=gist.get("created_at"),
                updated_at=gist.get("updated_at"),
                html_url=gist.get("html_url"),
                git_pull_url=gist.get("git_pull_url"),
                git_push_url=gist.get("git_push_url"),
                comments=gist.get("comments", 0),
                public=gist.get("public", False),
                owner=gist.get("owner", {}).get("login") if gist.get("owner") else None,
                files_count=len(gist.get("files", {})),
                # Include truncated file names as a preview
                files_preview=list(gist.get("files", {}).keys()),
                size=sum(f.get("size", 0) for f in gist.get("files", {}).values()),
            )
            for gist in all_gists
        ]

        self.dircache[""] = out
        return out

    @overload
    async def _ls(
        self, path: str, detail: Literal[True] = ..., **kwargs: Any
    ) -> list[GistInfo]: ...

    @overload
    async def _ls(self, path: str, detail: bool, **kwargs: Any) -> list[str]: ...

    async def _ls(
        self,
        path: str,
        detail: bool = True,
        **kwargs: Any,
    ) -> list[GistInfo] | list[str]:
        """List contents of path.

        Args:
            path: Path to list
            detail: Whether to include detailed information
            **kwargs: Additional arguments

        Returns:
            List of file/gist information or names

        Raises:
            FileNotFoundError: If path doesn't exist
        """
        path = self._strip_protocol(path or "")
        logger.debug("Listing path: %s (with gist_id: %s)", path, self.gist_id)

        # Different modes of operation
        if self.gist_id:
            # Single gist mode - always list files in this gist
            # Regardless of the path (root or specific file)
            results = await self._get_gist_file_list(self.gist_id)

            # If path is specified, filter to just that file
            if path:
                results = [f for f in results if f["name"] == path]
                if not results:
                    msg = f"File not found: {path}"
                    raise FileNotFoundError(msg)
        # User gists mode
        elif not path:
            # Root - list all gists
            results = await self._get_all_gists()
        else:
            # Specific gist - list its files
            parts = path.split("/", 1)
            gist_id = parts[0]

            try:
                results = await self._get_gist_file_list(gist_id)
            except FileNotFoundError:
                msg = f"Gist not found: {gist_id}"
                raise FileNotFoundError(msg)  # noqa: B904

            # If path includes a file, filter to just that file
            if len(parts) > 1 and parts[1]:
                file_name = parts[1]
                results = [f for f in results if f["name"] == file_name]
                if not results:
                    msg = f"File not found: {path}"
                    raise FileNotFoundError(msg)

        if detail:
            return results
        return [f["name"] for f in results]

    ls = sync_wrapper(_ls)

    async def _cat_file(
        self,
        path: str,
        start: int | None = None,
        end: int | None = None,
        **kwargs: Any,
    ) -> bytes:
        """Get contents of a file."""
        path = self._strip_protocol(path)
        if self.gist_id:
            gist_id = self.gist_id
            file_name = path
        else:  # Parse path into gist_id and file_name
            if "/" not in path:
                msg = f"Invalid file path: {path}"
                raise ValueError(msg)
            gist_id, file_name = path.split("/", 1)

        # Find file info in dircache
        files = await self._get_gist_file_list(gist_id)
        matches = [f for f in files if f["name"] == file_name]
        if not matches:
            msg = f"File not found: {path}"
            raise FileNotFoundError(msg)

        raw_url = matches[0].get("raw_url")
        if not raw_url:
            msg = f"No raw URL for file: {path}"
            raise FileNotFoundError(msg)

        session = await self.set_session()
        response = await session.get(raw_url)
        if response.status_code == 404:  # noqa: PLR2004
            msg = f"File not found: {path}"
            raise FileNotFoundError(msg)

        response.raise_for_status()
        content = response.content
        if start is not None or end is not None:
            start = start or 0
            end = min(end or len(content), len(content))
            content = content[start:end]
        return content

    cat_file = sync_wrapper(_cat_file)  # type: ignore

    async def _pipe_file(self, path: str, value: bytes, **kwargs: Any) -> None:
        """Write bytes to a file in a gist.

        Args:
            path: Path in format "gist_id/filename" or "filename" (for single gist mode)
            value: Content to write
            **kwargs: Additional keyword arguments
                gist_description: Optional description for new gists
                public: Whether the gist should be public (default: False)

        Raises:
            ValueError: If token is not provided for write operations
        """
        if not self.token:
            msg = "GitHub token is required for write operations"
            raise ValueError(msg)

        session = await self.set_session()
        path = self._strip_protocol(path)
        logger.debug("Writing to path: %s", path)
        # Parse the path into gist_id and filename
        if self.gist_id and "/" not in path:
            # Single gist mode with just a filename
            gist_id = self.gist_id
            filename = path
        else:
            # Path should include gist_id/filename
            if "/" not in path:
                msg = "Cannot create file without gist_id. Use 'gist_id/filename' format"
                raise ValueError(msg)
            gist_id, filename = path.split("/", 1)

        logger.debug("Resolved gist_id=%s, filename=%s", gist_id, filename)
        # Determine if we're updating an existing gist or creating a new one
        is_update = True
        try:
            await self._fetch_gist_metadata(gist_id)
        except FileNotFoundError:
            # Gist doesn't exist, create new one
            is_update = False
            logger.debug("Gist %s not found, will create new gist", gist_id)

        # Convert bytes to string content
        try:
            content = value.decode()
        except UnicodeDecodeError:
            # If content is binary, base64 encode it
            content = f"base64:{base64.b64encode(value).decode('ascii')}"

        files_data = {filename: {"content": content}}
        if is_update:
            # Update existing gist
            update_url = f"https://api.github.com/gists/{gist_id}"
            logger.debug("Updating existing gist: %s", update_url)
            data = {"files": files_data}
            response = await session.patch(update_url, json=data)
        else:
            # Create new gist
            create_url = "https://api.github.com/gists"
            logger.debug("Creating new gist: %s", create_url)
            description = kwargs.get("gist_description", "Gist created via GistFileSystem")
            public = kwargs.get("public", False)
            data = {"description": description, "public": public, "files": files_data}
            response = await session.post(create_url, json=data)

        if response.status_code >= 400:  # noqa: PLR2004
            logger.error("API error: %s %s", response.status_code, response.text)
            response.raise_for_status()

        # Invalidate cache for this gist
        self.dircache.pop(gist_id, None)

    pipe_file = sync_wrapper(_pipe_file)

    async def _rm_file(self, path: str, **kwargs: Any) -> None:
        """Delete a file from a gist.

        Args:
            path: Path in format "gist_id/filename" or "filename" (for single gist mode)
            **kwargs: Additional keyword arguments

        Raises:
            ValueError: If token is not provided for delete operations
        """
        if not self.token:
            msg = "GitHub token is required for delete operations"
            raise ValueError(msg)

        session = await self.set_session()
        path = self._strip_protocol(path)

        # Parse the path
        if self.gist_id and "/" not in path:
            gist_id = self.gist_id
            filename = path
        else:
            if "/" not in path:
                msg = "Cannot identify file without gist_id. Use 'gist_id/filename' format"
                raise ValueError(msg)
            gist_id, filename = path.split("/", 1)

        # To delete a file, we need to set its content to null in the API
        update_url = f"https://api.github.com/gists/{gist_id}"
        data = {"files": {filename: None}}  # Setting to null/None deletes the file

        response = await session.patch(update_url, json=data)
        response.raise_for_status()
        self.dircache.pop(gist_id, None)  # Invalidate cache for this gist

    rm_file = sync_wrapper(_rm_file)

    async def _rm(self, path: str, recursive: bool = False, **kwargs: Any) -> None:
        """Remove a file or entire gist.

        Args:
            path: Path to file or gist
            recursive: If True and path points to a gist, delete the entire gist
            **kwargs: Additional keyword arguments

        Raises:
            ValueError: If token is not provided for delete operations
        """
        if not self.token:
            msg = "GitHub token is required for delete operations"
            raise ValueError(msg)

        path = self._strip_protocol(path)
        # Determine if we're deleting a file or an entire gist
        if "/" in path or (self.gist_id and not path):
            # Path contains a filename or we're in single gist mode - delete file
            return await self._rm_file(path, **kwargs)

        # We're dealing with a gist ID directly
        gist_id = path if path else self.gist_id
        if not gist_id:
            msg = "No gist ID specified for deletion"
            raise ValueError(msg)

        if not recursive:
            msg = "Cannot delete a gist without recursive=True"
            raise ValueError(msg)

        # Delete the entire gist
        session = await self.set_session()
        delete_url = f"https://api.github.com/gists/{gist_id}"
        response = await session.delete(delete_url)
        response.raise_for_status()

        # Remove from cache
        self.dircache.pop(gist_id, None)
        self.dircache.pop("", None)  # Also invalidate root listing
        return None

    rm = sync_wrapper(_rm)

    async def _info(self, path: str, **kwargs: Any) -> GistInfo:
        """Get info about a path."""
        path = self._strip_protocol(path)
        if not path:
            return GistInfo(
                name="",
                type="directory",
                size=0,
                description="Root directory listing gists",
            )

        parts = path.split("/")
        if len(parts) == 1:
            if self.gist_id:
                # In single gist mode, this must be a file
                try:
                    files = await self._get_gist_file_list(self.gist_id)
                    matches = [f for f in files if f["name"] == path]
                    if not matches:
                        msg = f"File not found: {path}"
                        raise FileNotFoundError(msg)  # noqa: TRY301
                    file_info = matches[0]
                    return GistInfo(
                        name=file_info.get("name", path),
                        type="file",
                        size=file_info.get("size", 0),
                        description=file_info.get("description"),
                        filename=file_info.get("filename"),
                        raw_url=file_info.get("raw_url"),
                        language=file_info.get("language"),
                        content=file_info.get("content"),
                        truncated=file_info.get("truncated", False),
                    )
                except FileNotFoundError:
                    msg = f"File not found: {path}"
                    raise FileNotFoundError(msg)  # noqa: B904
            else:
                # In user gists mode, this is a gist ID
                try:
                    gists = await self._get_all_gists()
                    matches = [g for g in gists if g["name"] == parts[0]]
                    if not matches:
                        # Try to fetch the specific gist
                        try:
                            meta = await self._fetch_gist_metadata(parts[0])
                            return GistInfo(
                                name=parts[0],
                                type="directory",
                                description=meta.get("description", ""),
                                created_at=meta.get("created_at"),
                                updated_at=meta.get("updated_at"),
                                html_url=meta.get("html_url"),
                                git_pull_url=meta.get("git_pull_url"),
                                git_push_url=meta.get("git_push_url"),
                                comments=meta.get("comments", 0),
                                public=meta.get("public", False),
                                owner=meta.get("owner", {}).get("login")
                                if meta.get("owner")
                                else None,
                                files_count=len(meta.get("files", {})),
                                files_preview=list(meta.get("files", {}).keys()),
                                size=sum(f.get("size", 0) for f in meta.get("files", {}).values()),
                            )
                        except FileNotFoundError:
                            msg = f"Gist not found: {parts[0]}"
                            raise FileNotFoundError(msg)  # noqa: B904
                    file_info = matches[0]
                    return GistInfo(
                        name=file_info.get("name", parts[0]),
                        type="file",
                        size=file_info.get("size", 0),
                        description=file_info.get("description"),
                        filename=file_info.get("filename"),
                        raw_url=file_info.get("raw_url"),
                        language=file_info.get("language"),
                        content=file_info.get("content"),
                        truncated=file_info.get("truncated", False),
                    )
                except FileNotFoundError:
                    msg = f"Gist not found: {parts[0]}"
                    raise FileNotFoundError(msg)  # noqa: B904
        else:
            # This is a file within a gist
            gist_id = parts[0] if not self.gist_id else self.gist_id
            file_name = parts[1] if not self.gist_id else parts[0]

            try:
                files = await self._get_gist_file_list(gist_id)
                matches = [f for f in files if f["name"] == file_name]
                if not matches:
                    msg = f"File not found: {path}"
                    raise FileNotFoundError(msg)  # noqa: TRY301
                file_info = matches[0]
                return GistInfo(
                    name=file_info.get("name", file_name),
                    type="file",
                    size=file_info.get("size", 0),
                    description=file_info.get("description"),
                    filename=file_info.get("filename"),
                    raw_url=file_info.get("raw_url"),
                    language=file_info.get("language"),
                    content=file_info.get("content"),
                    truncated=file_info.get("truncated", False),
                )
            except FileNotFoundError:
                msg = f"File not found: {path}"
                raise FileNotFoundError(msg)  # noqa: B904

    info = sync_wrapper(_info)

    async def _exists(self, path: str, **kwargs: Any) -> bool:
        """Check if a path exists."""
        try:
            await self._info(path, **kwargs)
        except FileNotFoundError:
            return False
        else:
            return True

    exists = sync_wrapper(_exists)  # pyright: ignore

    async def _isdir(self, path: str, **kwargs: Any) -> bool:
        """Check if path is a directory."""
        if not path:
            return True  # Root is always a directory

        try:
            info = await self._info(path, **kwargs)
            return info["type"] == "directory"  # pyright: ignore[reportTypedDictNotRequiredAccess]
        except FileNotFoundError:
            return False

    isdir = sync_wrapper(_isdir)

    async def _isfile(self, path: str, **kwargs: Any) -> bool:
        """Check if path is a file."""
        try:
            info = await self._info(path, **kwargs)
            return info["type"] == "file"  # pyright: ignore[reportTypedDictNotRequiredAccess]
        except FileNotFoundError:
            return False

    isfile = sync_wrapper(_isfile)

    def _open(self, path: str, mode: str = "rb", **kwargs: Any) -> io.BytesIO | BufferedWriter:
        """Open a file.

        Args:
            path: Path to the file
            mode: File mode ('rb' for reading, 'wb' for writing)
            **kwargs: Additional arguments for write operations
                gist_description: Optional description for new gists
                public: Whether the gist should be public (default: False)

        Returns:
            File-like object for reading or writing

        Raises:
            ValueError: If token is not provided for write operations
            NotImplementedError: If mode is not supported
        """
        if "r" in mode:
            content = self.cat_file(path)
            assert isinstance(content, bytes)
            return io.BytesIO(content)
        if "w" in mode:
            if not self.token:
                msg = "GitHub token is required for write operations"
                raise ValueError(msg)

            buffer = io.BytesIO()
            return BufferedWriter(buffer, self, path, **kwargs)
        msg = f"Mode {mode} not supported"
        raise NotImplementedError(msg)

    def invalidate_cache(self, path: str | None = None) -> None:
        """Clear the cache."""
        if path is None:
            self.dircache.clear()
        else:
            path = self._strip_protocol(path)
            if self.gist_id:
                self.dircache.pop(self.gist_id, None)
            elif not path or path == "/":
                self.dircache.pop("", None)
            else:
                parts = path.split("/")
                if len(parts) >= 1:
                    self.dircache.pop(parts[0], None)


if __name__ == "__main__":
    import asyncio

    logging.basicConfig(level=logging.INFO)
    print(f"Environment GITHUB_TOKEN set: {'GITHUB_TOKEN' in os.environ}")
    fs = GistFileSystem(username="phil65")
    print("\nListing files with filesystem.ls():")

    async def main() -> None:
        upath = fs.get_upath("")
        async for p in upath.aiterdir():
            print(p)

    asyncio.run(main())
    # test_filename = "test_file2.py"
    # print(f"\nWriting to {test_filename}")
    # fs.pipe_file(test_filename, b"test content")
    # print("Write successful")
    # print("\nReading file:")
    # content = fs.cat_file(test_filename)
    # print(f"Content: {content}")
    #
    #
    # fs = GistFileSystem(username="phil65")
    # files = fs.ls("")



"""Skills-aware filesystem using WrapperFileSystem with info callback."""

from __future__ import annotations

import logging
from typing import TYPE_CHECKING, Any

import yaml

from upathtools import is_directory
from upathtools.filesystems.base.wrapper import WrapperFileSystem


if TYPE_CHECKING:
    from fsspec.asyn import AsyncFileSystem
    from fsspec.spec import AbstractFileSystem
    from upath.types import JoinablePathLike


logger = logging.getLogger(__name__)


async def _skill_info_callback(  # noqa: PLR0911
    info: dict[str, Any],
    fs: WrapperFileSystem,
) -> dict[str, Any]:
    """Enrich directory info with skill metadata if it contains SKILL.md."""
    if not await is_directory(fs, info["name"], entry_type=info.get("type")):
        return info

    path = info["name"]
    skill_path = f"{path.rstrip('/')}/SKILL.md"

    try:
        if not await fs.fs._exists(skill_path):
            return info
    except Exception:  # noqa: BLE001
        return info

    try:
        raw_content = await fs.fs._cat_file(skill_path)
        content = raw_content.decode("utf-8") if isinstance(raw_content, bytes) else raw_content
        if not content.startswith("---\n"):
            return info

        parts = content.split("---\n", 2)
        if len(parts) < 2:  # noqa: PLR2004
            return info

        frontmatter = parts[1].strip()
        metadata = yaml.safe_load(frontmatter) or {}

        logger.debug("Parsed skill metadata for %s: %s", path, metadata.get("name"))

        return {
            **info,
            "is_skill": True,
            "skill_name": metadata.get("name", ""),
            "skill_description": metadata.get("description", ""),
            "skill_version": metadata.get("version", ""),
            "skill_author": metadata.get("author", ""),
            "skill_tags": metadata.get("tags", []),
            "skill_dependencies": metadata.get("dependencies", []),
            "skill_metadata": metadata,
        }

    except yaml.YAMLError as e:
        logger.warning("Failed to parse YAML frontmatter in %s: %s", skill_path, e)
    except Exception as e:  # noqa: BLE001
        logger.debug("Could not parse skill metadata for %s: %s", path, e)

    return info


def create_skills_filesystem(
    wrapped_fs: AbstractFileSystem | AsyncFileSystem | JoinablePathLike,
    **storage_options: Any,
) -> WrapperFileSystem:
    """Create a filesystem that enriches directories with skill metadata.

    Args:
        wrapped_fs: Filesystem to wrap, or path to create filesystem from
        **storage_options: Additional options passed to wrapped filesystem

    Returns:
        WrapperFileSystem with skill metadata enrichment
    """
    from fsspec.asyn import AsyncFileSystem
    from fsspec.spec import AbstractFileSystem

    from upathtools.helpers import upath_to_fs

    if isinstance(wrapped_fs, AsyncFileSystem | AbstractFileSystem):
        fs = wrapped_fs
    else:
        fs = upath_to_fs(wrapped_fs, **storage_options)

    return WrapperFileSystem(fs=fs, info_callback=_skill_info_callback)


async def list_skills(fs: WrapperFileSystem, path: str = "/") -> list[dict[str, Any]]:
    """Get all skill directories under a path.

    Args:
        fs: A WrapperFileSystem (ideally created with create_skills_filesystem)
        path: Path to search for skills

    Returns:
        List of skill info dicts with path, name, description, and metadata
    """
    skills = []

    try:
        entries = await fs._ls(path, detail=True)

        for entry in entries:
            if entry.get("is_skill"):
                skills.append({
                    "path": entry["name"],
                    "name": entry.get("skill_name", ""),
                    "description": entry.get("skill_description", ""),
                    "metadata": entry.get("skill_metadata", {}),
                })
            elif await is_directory(fs, entry["name"], entry_type=entry.get("type")):
                subskills = await list_skills(fs, entry["name"])
                skills.extend(subskills)

    except Exception as e:  # noqa: BLE001
        logger.warning("Failed to list skills in %s: %s", path, e)

    return skills


if __name__ == "__main__":
    import asyncio

    fs = create_skills_filesystem("file:///home/phil65/dev/oss/upathtools/.claude/skills/")

    async def main() -> None:
        skills = await list_skills(fs)
        print(skills)

    asyncio.run(main())



"""Logging configuration for upathtools."""

from __future__ import annotations

import logging


LogLevel = int | str


def get_logger(name: str, log_level: LogLevel | None = None) -> logging.Logger:
    """Get a logger for the given name.

    Args:
        name: The name of the logger.
              Will be prefixed with 'upathtools'
        log_level: The logging level to set for the logger

    Returns:
        A logger instance
    """
    logger = logging.getLogger(f"upathtools.{name}")
    if log_level is not None:
        logger.setLevel(log_level)
    return logger



"""Filesystem implementation for browsing SQLite databases."""

from __future__ import annotations

import asyncio
import contextlib
import csv
import io
from pathlib import Path
import tempfile
from typing import TYPE_CHECKING, Any, ClassVar, Literal, Required, overload

import fsspec

from upathtools.async_helpers import sync_wrapper
from upathtools.filesystems.base import (
    BaseAsyncFileFileSystem,
    BaseUPath,
    FileInfo,
    GrepMatch,
    ProbeResult,
)


if TYPE_CHECKING:
    from collections.abc import Sequence

    from fsspec.asyn import AsyncFileSystem
    from sqlalchemy.ext.asyncio import AsyncEngine


class SqliteInfo(FileInfo, total=False):
    """Info dict for SQLite filesystem paths."""

    size: Required[int]
    table_type: str


class SqlitePath(BaseUPath[SqliteInfo]):
    """UPath implementation for browsing SQLite databases."""

    __slots__ = ()

    def iterdir(self):
        if not self.is_dir():
            raise NotADirectoryError(str(self))
        yield from super().iterdir()


class SqliteFileSystem(BaseAsyncFileFileSystem[SqlitePath, SqliteInfo]):
    """Filesystem for browsing SQLite databases."""

    protocol = "sqlite"
    upath_cls = SqlitePath
    supported_extensions: ClassVar[frozenset[str]] = frozenset({"db", "sqlite", "sqlite3"})
    priority: ClassVar[int] = 50

    # SQLite magic bytes: "SQLite format 3\x00"
    SQLITE_MAGIC = b"SQLite format 3\x00"

    @classmethod
    def probe_content(cls, content: bytes, extension: str = "") -> ProbeResult:
        """Probe content to check if it's a SQLite database.

        Checks for SQLite magic bytes at the start of the file.
        """
        if len(content) >= 16 and content[:16] == cls.SQLITE_MAGIC:  # noqa: PLR2004
            return ProbeResult.SUPPORTED
        # Extension matches but no magic bytes - might be empty or corrupted
        if cls.supports_extension(extension):
            return ProbeResult.MAYBE
        return ProbeResult.UNSUPPORTED

    @classmethod
    def get_probe_size(cls) -> int | None:
        """Only need first 16 bytes to check SQLite magic."""
        return 16

    def __init__(
        self,
        db_path: str = "",
        target_protocol: str | None = None,
        target_options: dict[str, Any] | None = None,
        parent_fs: AsyncFileSystem | None = None,
        parent_path: str | None = None,
        **kwargs: Any,
    ) -> None:
        """Initialize the filesystem.

        Args:
            db_path: Path to SQLite database file (local path or temp file)
            target_protocol: Protocol for source database file
            target_options: Options for target protocol
            parent_fs: Parent filesystem for read-write access (optional)
            parent_path: Path within parent filesystem (optional)
            **kwargs: Additional filesystem options
        """
        super().__init__(**kwargs)
        self.db_path = db_path
        self.target_protocol = target_protocol
        self.target_options = target_options or {}
        self._parent_fs = parent_fs
        self._parent_path = parent_path
        self._engine: AsyncEngine | None = None
        self._temp_file: str | None = None
        self._is_temp_copy = False

    @classmethod
    def from_file(
        cls,
        path: str,
        target_protocol: str | None = None,
        target_options: dict[str, Any] | None = None,
        **kwargs: Any,
    ) -> SqliteFileSystem:
        """Create filesystem instance from a SQLite database file path."""
        return cls(
            db_path=path, target_protocol=target_protocol, target_options=target_options, **kwargs
        )

    @classmethod
    async def from_filesystem_async(
        cls,
        path: str,
        fs: AsyncFileSystem,
        **kwargs: Any,
    ) -> SqliteFileSystem:
        """Create filesystem instance with parent filesystem access.

        For local files, uses the file directly. For remote files,
        creates a temp copy but retains parent fs reference for potential
        write-back operations.
        """
        from fsspec.implementations.local import LocalFileSystem

        # Check if parent fs is local - if so, use file directly
        underlying = getattr(fs, "fs", None)
        if isinstance(fs, LocalFileSystem) or isinstance(underlying, LocalFileSystem):
            return cls(db_path=path, parent_fs=fs, parent_path=path, **kwargs)

        # Remote file - download to temp, keep parent reference for write-back
        content = await fs._cat_file(path)
        with tempfile.NamedTemporaryFile(delete=False, suffix=".db") as tmp:
            if isinstance(content, str):
                content = content.encode()
            tmp.write(content)
            tmp_name = tmp.name

        instance = cls(db_path=tmp_name, parent_fs=fs, parent_path=path, **kwargs)
        instance._temp_file = tmp_name
        instance._is_temp_copy = True
        return instance

    @classmethod
    def from_content(cls, content: bytes, **kwargs: Any) -> SqliteFileSystem:
        """Create filesystem instance from raw SQLite database content.

        Note: This creates a read-only temp copy. Use from_filesystem_async
        for read-write access to remote databases.
        """
        with tempfile.NamedTemporaryFile(delete=False, suffix=".db") as tmp:
            tmp.write(content)
            tmp_name = tmp.name
        instance = cls(db_path=tmp_name, **kwargs)
        instance._temp_file = tmp_name
        instance._is_temp_copy = True
        return instance

    async def sync_to_parent(self) -> None:
        """Write temp database back to parent filesystem.

        Only applicable when created via from_filesystem_async with a remote fs.
        """
        if not self._is_temp_copy or self._parent_fs is None or self._parent_path is None:
            return

        # Use asyncio executor to avoid blocking
        content = await asyncio.get_event_loop().run_in_executor(
            None, Path(self.db_path).read_bytes
        )

        try:
            await self._parent_fs._pipe_file(self._parent_path, content)
        except AttributeError:
            self._parent_fs.pipe_file(self._parent_path, content)

    @staticmethod
    def _get_kwargs_from_urls(path: str) -> dict[str, Any]:
        path = path.removeprefix("sqlite://")
        return {"db_path": path}

    async def _get_engine(self) -> AsyncEngine:
        """Get or create async SQLAlchemy engine."""
        from sqlalchemy.ext.asyncio import create_async_engine

        if self._engine is not None:
            return self._engine

        if self.target_protocol:
            # Download remote DB to temp file
            with (
                fsspec.open(
                    self.db_path,
                    protocol=self.target_protocol,
                    **self.target_options,
                ) as f,
                tempfile.NamedTemporaryFile(delete=False) as tmp,
            ):
                tmp.write(f.read())  # type: ignore[reportArgumentType]
                self._temp_file = tmp.name
                db_url = f"sqlite+aiosqlite:///{tmp.name}"
        else:
            db_url = f"sqlite+aiosqlite:///{self.db_path}"

        self._engine = create_async_engine(db_url)
        return self._engine

    @overload
    async def _ls(
        self,
        path: str,
        detail: Literal[True] = ...,
        **kwargs: Any,
    ) -> list[SqliteInfo]: ...

    @overload
    async def _ls(self, path: str, detail: bool, **kwargs: Any) -> list[str]: ...

    async def _ls(
        self,
        path: str,
        detail: bool = True,
        **kwargs: Any,
    ) -> Sequence[str | SqliteInfo]:
        """List database tables and views."""
        from sqlalchemy import text

        engine = await self._get_engine()
        async with engine.begin() as conn:
            # Get table names
            result = await conn.execute(
                text("""
                SELECT name, type FROM sqlite_master
                WHERE type IN ('table', 'view')
                ORDER BY name
                """)
            )

            items = []
            for row in result:  # type: ignore[reportAttributeAccessIssue]
                if detail:
                    # Could add COUNT(*) query if needed
                    item = SqliteInfo(name=row.name, type="file", size=0, table_type=row.type)  # type: ignore[reportAttributeAccessIssue]
                    items.append(item)
                else:
                    items.append(row.name)  # type: ignore[reportAttributeAccessIssue]

        return items

    ls = sync_wrapper(_ls)  # pyright: ignore[reportAssignmentType]

    async def _cat_file(
        self, path: str, start: int | None = None, end: int | None = None, **kwargs: Any
    ) -> bytes:
        """Get table data as CSV."""
        from sqlalchemy import text

        engine = await self._get_engine()
        path = self._strip_protocol(path).strip("/")  # type: ignore[reportAttributeAccessIssue]

        if not path:
            msg = "Cannot cat root directory"
            raise IsADirectoryError(msg)

        # Handle special files
        if path == ".schema":
            return await self._get_schema()
        if path.endswith(".schema"):
            table_name = path.removesuffix(".schema")
            return await self._get_table_schema(table_name)

        # Regular table data
        async with engine.begin() as conn:
            result = await conn.execute(text(f"SELECT * FROM `{path}`"))
            rows = result.fetchall()
            columns = result.keys()

            # Convert to CSV
            output = io.StringIO()
            writer = csv.writer(output)
            writer.writerow(columns)
            writer.writerows(rows)

            content = output.getvalue().encode()

            # Handle byte range if specified
            if start is not None or end is not None:
                start = start or 0
                end = end or len(content)
                content = content[start:end]

            return content

    cat_file = sync_wrapper(_cat_file)  # pyright: ignore[reportAssignmentType]

    async def _get_schema(self) -> bytes:
        """Get full database schema."""
        from sqlalchemy import text

        engine = await self._get_engine()

        async with engine.begin() as conn:
            result = await conn.execute(
                text("""
                    SELECT sql FROM sqlite_master
                    WHERE sql IS NOT NULL
                    ORDER BY type, name
                """)
            )

            schema_lines = [
                row.sql + ";"  # type: ignore[reportAttributeAccessIssue]
                for row in result  # type: ignore[reportAttributeAccessIssue]
                if row.sql  # type: ignore[reportAttributeAccessIssue]
            ]

            return "\n".join(schema_lines).encode()

    async def _get_table_schema(self, table_name: str) -> bytes:
        """Get schema for specific table."""
        from sqlalchemy import text

        engine = await self._get_engine()

        async with engine.begin() as conn:
            result = await conn.execute(
                text("SELECT sql FROM sqlite_master WHERE name = :name AND sql IS NOT NULL"),
                {"name": table_name},
            )
            row = result.fetchone()

            if row and row.sql:  # type: ignore[reportAttributeAccessIssue]
                return (row.sql + ";").encode()  # type: ignore[reportAttributeAccessIssue]

            msg = f"Table {table_name} not found"
            raise FileNotFoundError(msg)

    async def _info(self, path: str, **kwargs: Any) -> SqliteInfo:
        """Get info about database objects."""
        from sqlalchemy import text

        engine = await self._get_engine()
        path = self._strip_protocol(path).strip("/")  # type: ignore[reportAttributeAccessIssue]

        if not path or path == "/":
            # Root directory info
            return SqliteInfo(
                name="root",
                type="directory",
                size=0,
            )

        # Handle special files
        if path == ".schema":
            schema_content = await self._get_schema()
            return SqliteInfo(
                name=".schema",
                type="file",
                size=len(schema_content),
            )
        if path.endswith(".schema"):
            table_name = path.removesuffix(".schema")
            schema_content = await self._get_table_schema(table_name)
            return SqliteInfo(
                name=path,
                type="file",
                size=len(schema_content),
            )

        # Regular table info
        async with engine.begin() as conn:
            # Check if table exists
            result = await conn.execute(
                text("SELECT type FROM sqlite_master WHERE name = :name"),
                {"name": path},
            )
            row = result.fetchone()

            if not row:
                msg = f"Table {path} not found"
                raise FileNotFoundError(msg)

            # Get row count
            count_result = await conn.execute(text(f"SELECT COUNT(*) FROM `{path}`"))
            count = count_result.scalar()

            return SqliteInfo(
                name=path,
                type="file",
                size=count or 0,
                table_type=row.type,  # type: ignore[reportAttributeAccessIssue]
            )

    info = sync_wrapper(_info)

    async def _exists(self, path: str, **kwargs: Any) -> bool:
        """Check if table or view exists."""
        try:
            await self._info(path)
        except FileNotFoundError:
            return False
        else:
            return True

    exists = sync_wrapper(_exists)  # pyright: ignore[reportAssignmentType]

    async def _isdir(self, path: str, **kwargs: Any) -> bool:
        """Check if path is a directory."""
        path = self._strip_protocol(path).strip("/")  # type: ignore[reportAttributeAccessIssue]
        return not path or path == "/"

    isdir = sync_wrapper(_isdir)

    async def _isfile(self, path: str, **kwargs: Any) -> bool:
        """Check if path is a file."""
        return await self._exists(path) and not await self._isdir(path)

    isfile = sync_wrapper(_isfile)

    @overload
    async def _glob(
        self,
        path: str,
        maxdepth: int | None = None,
        *,
        detail: Literal[False] = False,
        **kwargs: Any,
    ) -> list[str]: ...

    @overload
    async def _glob(
        self,
        path: str,
        maxdepth: int | None = None,
        *,
        detail: Literal[True],
        **kwargs: Any,
    ) -> dict[str, SqliteInfo]: ...

    async def _glob(
        self,
        path: str,
        maxdepth: int | None = None,
        *,
        detail: bool = False,
        **kwargs: Any,
    ) -> list[str] | dict[str, SqliteInfo]:
        """Glob for tables using SQL LIKE pattern matching.

        Converts glob patterns to SQL LIKE patterns for efficient matching.
        """
        import fnmatch

        from sqlalchemy import text

        path = self._strip_protocol(path).strip("/")  # type: ignore[reportAttributeAccessIssue]

        # Check for glob magic characters
        glob_chars = {"*", "?", "["}
        if not any(c in path for c in glob_chars):
            if await self._exists(path):
                if detail:
                    info = await self._info(path)
                    return {path: info}
                return [path]
            return {} if detail else []

        engine = await self._get_engine()

        async with engine.begin() as conn:
            result = await conn.execute(
                text("""
                SELECT name FROM sqlite_master
                WHERE type IN ('table', 'view')
                ORDER BY name
                """)
            )
            all_tables = [row.name for row in result]  # type: ignore[reportAttributeAccessIssue]

        # Use fnmatch for glob pattern matching
        matches = [t for t in all_tables if fnmatch.fnmatch(t, path)]

        if not detail:
            return matches

        # Return barebone info dicts
        return {t: SqliteInfo(name=t, type="file", size=0) for t in matches}

    glob = sync_wrapper(_glob)  # pyright: ignore[reportAssignmentType]

    async def _grep(
        self,
        path: str,
        pattern: str,
        *,
        max_count: int | None = None,
        case_sensitive: bool | None = None,
        hidden: bool = False,
        no_ignore: bool = False,
        globs: list[str] | None = None,
        context_before: int | None = None,
        context_after: int | None = None,
        multiline: bool = False,
    ) -> list[GrepMatch]:
        """Search table contents using SQL LIKE.

        Searches all text columns in the specified table for the pattern.
        Much faster than reading CSV and searching in Python.
        """
        from sqlalchemy import text

        path = self._strip_protocol(path).strip("/")  # type: ignore[reportAttributeAccessIssue]

        if not path or path == "/":
            msg = "Cannot grep root directory"
            raise IsADirectoryError(msg)

        engine = await self._get_engine()

        async with engine.begin() as conn:
            # Get column info for the table
            col_result = await conn.execute(text(f"PRAGMA table_info(`{path}`)"))
            columns = [row[1] for row in col_result]  # column name is at index 1

            if not columns:
                msg = f"Table {path} not found"
                raise FileNotFoundError(msg)

            # Build WHERE clause to search all text columns
            like_pattern = f"%{pattern}%"
            if case_sensitive:
                conditions = " OR ".join(f"CAST(`{col}` AS TEXT) LIKE :pattern" for col in columns)
            else:
                conditions = " OR ".join(
                    f"LOWER(CAST(`{col}` AS TEXT)) LIKE LOWER(:pattern)" for col in columns
                )

            query = f"SELECT rowid, * FROM `{path}` WHERE {conditions}"
            result = await conn.execute(text(query), {"pattern": like_pattern})

            matches: list[GrepMatch] = []
            check_pattern = pattern if case_sensitive else pattern.lower()
            for row in result:
                row_dict = dict(row._mapping)
                rowid = row_dict.pop("rowid", None)
                # Find which columns matched
                for val in row_dict.values():
                    val_str = str(val) if val is not None else None
                    if val_str is not None and check_pattern in (
                        val_str if case_sensitive else val_str.lower()
                    ):
                        matches.append(
                            GrepMatch(
                                path=path,
                                line_number=rowid or 0,
                                text=val_str,
                            )
                        )
                        if max_count is not None and len(matches) >= max_count:
                            return matches

            return matches

    grep = sync_wrapper(_grep)

    def _open(self, path: str, mode: str = "rb", **kwargs: Any) -> Any:
        """Provide file-like access to table data."""
        if "w" in mode or "a" in mode:
            msg = "Write mode not supported"
            raise NotImplementedError(msg)
        try:
            loop = asyncio.get_event_loop()
        except RuntimeError:
            loop = asyncio.new_event_loop()
            asyncio.set_event_loop(loop)

        content = loop.run_until_complete(self._cat_file(path))
        return io.BytesIO(content)

    async def _close(self) -> None:
        """Close the database engine and clean up resources."""
        if self._engine is not None:
            await self._engine.dispose()
            self._engine = None

    close = sync_wrapper(_close)

    def __del__(self) -> None:
        """Clean up resources."""
        if self._temp_file:
            with contextlib.suppress(OSError):
                Path(self._temp_file).unlink()


if __name__ == "__main__":
    import sqlite3
    import tempfile

    async def demo():
        # Create a demo database
        with tempfile.NamedTemporaryFile(suffix=".db", delete=False) as f:
            demo_db = f.name

        # Create some test data
        conn = sqlite3.connect(demo_db)
        conn.execute("""
            CREATE TABLE users (
                id INTEGER PRIMARY KEY,
                name TEXT NOT NULL,
                email TEXT UNIQUE
            )
        """)
        conn.execute("""
            CREATE TABLE orders (
                id INTEGER PRIMARY KEY,
                user_id INTEGER,
                amount REAL,
                FOREIGN KEY (user_id) REFERENCES users (id)
            )
        """)
        conn.executemany(
            "INSERT INTO users (name, email) VALUES (?, ?)",
            [("Alice", "alice@example.com"), ("Bob", "bob@example.com")],
        )
        conn.executemany(
            "INSERT INTO orders (user_id, amount) VALUES (?, ?)",
            [(1, 100.0), (1, 200.0), (2, 50.0)],
        )
        conn.commit()
        conn.close()

        # Create filesystem
        fs = SqliteFileSystem(demo_db)

        # List tables
        print("\nTables:")
        tables = await fs._ls("/", detail=True)
        for table in tables:
            print(f"- {table['name']} ({table.get('table_type')})")

        # Read table data
        print("\nUsers table:")
        users_data = await fs._cat_file("users")
        print(users_data.decode())

        # Get schema
        print("\nDatabase schema:")
        schema = await fs._cat_file(".schema")
        print(schema.decode())

        # Clean up
        from pathlib import Path

        Path(demo_db).unlink()

    asyncio.run(demo())



"""Filesystem implementation for browsing markdown documents by header hierarchy."""

from __future__ import annotations

import re
import sys
from typing import TYPE_CHECKING, Any, ClassVar, Literal, Required, overload

import fsspec

from upathtools.filesystems.base import BaseAsyncFileFileSystem, BaseUPath, FileInfo, ProbeResult


if TYPE_CHECKING:
    from collections.abc import Sequence


class MarkdownInfo(FileInfo, total=False):
    """Info dict for Markdown filesystem paths."""

    size: Required[int]
    level: Required[int]


class MarkdownNode:
    """Represents a markdown header and its content."""

    def __init__(
        self,
        title: str,
        level: int,
        content: str = "",
        children: dict[str, MarkdownNode] | None = None,
    ) -> None:
        """Initialize a markdown node.

        Args:
            title: Header text
            level: Header level (number of #)
            content: Content belonging to this header
            children: Child nodes (sub-headers)
        """
        self.title = title
        self.level = level
        self.content = content
        self.children = children or {}

    def is_dir(self) -> bool:
        """Check if node should be treated as directory."""
        return bool(self.children)

    def get_size(self) -> int:
        """Get size of node's content."""
        return len(self.content.encode())


class MarkdownPath(BaseUPath[MarkdownInfo]):
    """UPath implementation for browsing markdown documents."""

    __slots__ = ()

    def iterdir(self):
        if not self.is_dir():
            raise NotADirectoryError(str(self))
        yield from super().iterdir()


class MarkdownFileSystem(BaseAsyncFileFileSystem[MarkdownPath, MarkdownInfo]):
    """Filesystem for browsing markdown documents by header hierarchy."""

    protocol = "md"
    upath_cls = MarkdownPath
    supported_extensions: ClassVar[frozenset[str]] = frozenset({"md", "markdown", "mdown", "mkd"})
    priority: ClassVar[int] = 50

    @classmethod
    def probe_content(cls, content: bytes, extension: str = "") -> ProbeResult:
        """Probe content to check if it's a markdown document.

        Looks for markdown indicators like headers (#) or common markdown syntax.
        """
        if not cls.supports_extension(extension):
            return ProbeResult.UNSUPPORTED

        try:
            text = content.decode("utf-8")
            # Check for markdown headers
            header_pattern = re.compile(r"^#{1,6}\s+.+$", re.MULTILINE)
            if header_pattern.search(text):
                return ProbeResult.SUPPORTED

            # Check for other common markdown patterns
            markdown_patterns = [
                r"^\s*[-*+]\s+",  # Unordered lists
                r"^\s*\d+\.\s+",  # Ordered lists
                r"\[.+\]\(.+\)",  # Links
                r"```",  # Code blocks
                r"^\s*>\s+",  # Blockquotes
            ]
            for pattern in markdown_patterns:
                if re.search(pattern, text, re.MULTILINE):
                    return ProbeResult.MAYBE

            # Extension matches but no clear markdown indicators
        except Exception:  # noqa: BLE001
            return ProbeResult.UNSUPPORTED
        else:
            return ProbeResult.MAYBE

    def __init__(
        self,
        fo: str = "",
        target_protocol: str | None = None,
        target_options: dict[str, Any] | None = None,
        **kwargs: Any,
    ) -> None:
        """Initialize the filesystem.

        Args:
            fo: Path to markdown file
            target_protocol: Protocol for source file
            target_options: Options for target protocol
            **kwargs: Additional filesystem options
        """
        super().__init__(**kwargs)
        self.path = fo
        self.target_protocol = target_protocol
        self.target_options = target_options or {}
        self._content: str | None = None
        self._root: MarkdownNode | None = None

    @classmethod
    def from_content(
        cls,
        content: bytes,
        **kwargs: Any,
    ) -> MarkdownFileSystem:
        """Create filesystem instance from raw markdown content.

        Args:
            content: Raw markdown content as bytes.
            **kwargs: Additional filesystem options.

        Returns:
            Configured filesystem instance with pre-loaded content.
        """
        fs = cls(fo="<content>", **kwargs)
        fs._content = content.decode("utf-8")
        fs._parse_content()
        return fs

    @classmethod
    def from_file(
        cls,
        path: str,
        target_protocol: str | None = None,
        target_options: dict[str, Any] | None = None,
        **kwargs: Any,
    ) -> MarkdownFileSystem:
        """Create filesystem instance from a markdown file path."""
        return cls(
            fo=path, target_protocol=target_protocol, target_options=target_options, **kwargs
        )

    @staticmethod
    def _get_kwargs_from_urls(path: str) -> dict[str, Any]:
        path = path.removeprefix("md://")
        return {"fo": path}

    def _load(self) -> None:
        """Load and parse the markdown file if not already loaded."""
        if self._content is not None:
            return

        with fsspec.open(
            self.path,
            mode="r",
            protocol=self.target_protocol,
            **self.target_options,
        ) as f:
            self._content = f.read()  # pyright: ignore

        self._parse_content()

    def _parse_content(self) -> None:
        """Parse markdown content into node hierarchy."""
        if not self._content:
            self._root = MarkdownNode("root", 0)
            return

        lines = self._content.splitlines()
        header_pattern = re.compile(r"^(#+)\s+(.+)$")

        # First pass: find minimum header level
        min_level = sys.maxsize
        for line in lines:
            if match := header_pattern.match(line):
                level = len(match.group(1))
                min_level = min(level, min_level)

        # If no headers found, or only higher levels, default to 1
        min_level = min(min_level, 1)

        current_path: list[MarkdownNode] = []
        current_content: list[str] = []

        # Create root node
        self._root = MarkdownNode("root", 0)
        current_path.append(self._root)

        for line in lines:
            if match := header_pattern.match(line):
                # Process accumulated content
                if current_content and current_path:
                    current_path[-1].content = "\n".join(current_content)
                current_content = []

                # Process header
                actual_level = len(match.group(1))
                # Adjust level relative to minimum found
                level = actual_level - min_level + 1
                title = match.group(2).strip()
                node = MarkdownNode(title, level)

                # Find appropriate parent
                while current_path and current_path[-1].level >= level:
                    current_path.pop()

                # Skip if we lost the path (header level too deep)
                if not current_path:
                    continue

                # Add to parent's children
                current_path[-1].children[title] = node
                current_path.append(node)
            else:
                current_content.append(line)

        # Handle remaining content
        if current_content and current_path:
            current_path[-1].content = "\n".join(current_content)

    def _get_node(self, path: str) -> MarkdownNode:
        """Get node at path.

        Args:
            path: Path to node

        Returns:
            MarkdownNode at path

        Raises:
            FileNotFoundError: If path doesn't exist
        """
        self._load()
        assert self._root is not None

        if not path or path == "/":
            return self._root

        current = self._root
        parts = self._strip_protocol(path).strip("/").split("/")  # pyright: ignore[reportAttributeAccessIssue]

        for part in parts:
            if part not in current.children:
                msg = f"Section not found: {path}"
                raise FileNotFoundError(msg)
            current = current.children[part]

        return current

    @overload
    async def _ls(
        self, path: str, detail: Literal[True] = ..., **kwargs: Any
    ) -> list[MarkdownInfo]: ...

    @overload
    async def _ls(self, path: str, detail: bool, **kwargs: Any) -> list[str]: ...

    async def _ls(
        self, path: str, detail: bool = True, **kwargs: Any
    ) -> Sequence[str | MarkdownInfo]:
        """List contents of a path."""
        node = self._get_node(path)

        if not detail:
            return list(node.children)

        return [
            MarkdownInfo(
                name=name,
                size=child.get_size(),
                type="directory" if child.is_dir() else "file",
                level=child.level,
            )
            for name, child in node.children.items()
        ]

    async def _cat_file(
        self, path: str, start: int | None = None, end: int | None = None, **kwargs: Any
    ) -> bytes:
        """Get section content including header and all content up to next header."""
        data = self._get_content(path)
        return data[start:end]

    def _get_content(self, path: str) -> bytes:
        """Get the raw content bytes for a path."""
        node = self._get_node(path)

        # Skip header generation for root
        if node is self._root:
            return node.content.encode()

        # Build section content
        lines = []
        # Add header
        lines.append(f"{'#' * node.level} {node.title}")

        # Add content if any
        if node.content:
            lines.append(node.content)

        return "\n".join(lines).encode()

    async def _info(self, path: str, **kwargs: Any) -> MarkdownInfo:
        """Get info about a path."""
        node = self._get_node(path)
        name = "root" if not path or path == "/" else path.rsplit("/", maxsplit=1)[-1]

        return MarkdownInfo(
            name=name,
            size=node.get_size(),
            type="directory" if node.is_dir() else "file",
            level=node.level,
        )

    async def _isdir(self, path: str) -> bool:
        """Check if path is a directory (has children)."""
        try:
            node = self._get_node(path)
            return node.is_dir()
        except FileNotFoundError:
            return False


if __name__ == "__main__":
    # Example markdown content
    EXAMPLE = """\
# Section 1
Content 1
### Skipped Section
Skipped content
## Section 1.1
Content 1.1
## Section 1.2
Content 1.2
# Section 2
Content 2
## Section 2.1
Content 2.1
"""

    import tempfile

    # Create temporary file
    with tempfile.NamedTemporaryFile(mode="w", suffix=".md", delete=False) as f:
        f.write(EXAMPLE)
        f.flush()

        # Create filesystem
        fs = MarkdownFileSystem(f.name)

        # List root
        print("\nRoot sections:")
        for item in fs.ls("/", detail=True):
            print(f"- {item['name']} ({item['type']})")

        # List subsections
        print("\nSubsections of Section 1:")
        for item in fs.ls("Section 1", detail=True):
            print(f"- {item['name']} ({item['type']})")

        # Read content
        print("\nContent of Section 1.1:")
        print(fs.cat("Section 1/Section 1.1").decode())



"""Remote filesystems."""

from upathtools.filesystems.remote_filesystems.github_fs import (
    GithubFileSystem,
    GithubInfo,
    GithubPath,
)
from upathtools.filesystems.remote_filesystems.gitlab_fs import (
    GitLabFileSystem,
    GitLabInfo,
    GitLabPath,
)
from upathtools.filesystems.remote_filesystems.linear_fs import (
    LinearCommentInfo,
    LinearIssueFileSystem,
    LinearIssueInfo,
    LinearIssuePath,
    LinearProjectInfo,
)

__all__ = [
    "GitLabFileSystem",
    "GitLabInfo",
    "GitLabPath",
    "GithubFileSystem",
    "GithubInfo",
    "GithubPath",
    "LinearCommentInfo",
    "LinearIssueFileSystem",
    "LinearIssueInfo",
    "LinearIssuePath",
    "LinearProjectInfo",
]
